# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  OptimizationControlMechanism *************************************************


"""

Overview
--------

An OptimizationControlMechanism is a `ControlMechanism <ControlMechanism>` that uses an `OptimizationFunction` to find
an optimal `control_allocation <ControlMechanism.control_allocation>` for a given `state
<OptimizationControlMechanism_State>`. The `OptimizationFunction` uses the OptimizationControlMechanism's
`evaluation_function` <OptimizationControlMechanism.evalutate_function>` to evaluate `control_allocation
<ControlMechanism.control_allocation>` samples, and then implements the one that yields the best predicted result.
The result returned by the `evaluation_function` <OptimizationControlMechanism.evalutate_function>` is ordinarily
the `net_outcome <ControlMechanism.net_outcome>` computed by the OptimizationControlMechanism for the `Composition`
(or part of one) that it controls, and its `ObjectiveFunction` seeks to maximize this, which corresponds to
maximizing the Expected Value of Control, as described below.

.. _OptimizationControlMechanism_EVC:

**Expected Value of Control**

The `net_outcome <ControlMechanism.net_outcome>` of an OptmizationControlMechanism, like any `ControlMechanism`
is computed as the difference between the `outcome <ControlMechanism.outcome>` computed by its `objective_mechanism
<ControlMechanism.objective_mechanism>` and the `costs <ControlMechanism.costs>` of its `control_signals
<ControlMechanism.control_signals>` for a given `state <OptimizationControlMechanism_State>` (i.e.,
set of `feature_values <OptimizationControlMechanism.feature_values>` and `control_allocation
<ControlMechanism.control_allocation>`.  If the `outcome <ControlMechanism.outcome>` is configured to measure the
value of processing (e.g., reward received, time taken to respond, or a combination of these, etc.),
and the `OptimizationFunction` assigned as the OptimizationControlMechanism's `function
<OptimizationControlMechanism.function>` is configured find the `control_allocation
<ControlMechanism.control_allocation>` that maximizes its `net_outcome <ControlMechanism.net_outcome>`,
then the OptimizationControlMechanism is said to be maximizing the `Expected Value of Control (EVC)
<https://www.ncbi.nlm.nih.gov/pubmed/23889930>`_.  That is, it implements a cost-benefit analysis that
weighs the `costs <ControlMechanism.costs>` of the ControlSignal `values <ControlSignal.value>` specified by a
`control_allocation <ControlMechanism.control_allocation>` against the `outcome <ControlMechanism.outcome>` expected
to result from it.  The costs are computed based on the `cost_options <ControlSignal.cost_options>` specified for
each of the OptimizationControlMechanism's `control_signals <ControlMechanism.control_signals>` and its
`combine_costs <ControlMechanism.combine_costs>` function.  The EVC is determined by its `compute_net_outcome
<ControlMechanism.compute_net_outcome>` function (assigned to its `net_outcome <ControlMechanism.net_outcome>`
attribute), which is computed for a given `state <OptimizationControlMechanism_State>` by the
OptimizationControlMechanism's `evaluation_function <OptimizationControlMechanism.evalutation_function>`.

COMMENT:
The table `below <OptimizationControlMechanism_Examples>` lists different
parameterizations of OptimizationControlMechanism that implement various models of EVC Optimization.

### THROUGHOUT DOCUMENT, REWORD AS "optimizing control_allocation" RATHER THAN "maximizing" / "greatest"

COMMENT

.. _OptimizationControlMechanism_Agent_Representation_Types:

**Agent Representation and Types of Optimization**

The defining characteristic of an OptimizationControlMechanism is its `agent representation
<OptimizationControlMechanism_Agent_Rep>`, that is used to determine the `net_outcome <ControlMechanism.net_outcome>
for a given `state <OptimizationControlMechanism_State>` and find the `control_allocation
<ControlMechanism.control_allocation>` that optimizes this.  The `agent_rep <OptimizationControlMechanism.agent_rep>`
can be either the `Composition` to which the OptimizationControlMechanism belongs (and controls) or another one that
is used to estimate the `net_outcome <ControlMechanism.net_outcome>` for that Composition.  This distinction
corresponds closely to the distinction between *model-based* and *model-free* optimization in the `machine learning
<HTML REF>`_ and `cognitive neuroscience <https://www.nature.com/articles/nn1560>`_ literatures, as described below.

.. _OptimizationControlMechanism_Model_Free_Model_Based:

.. _OptimizationControlMechanism_Model_Free:

*Model-Free Optimization*

This is achieved by assigning as the `agent_rep  <OptimizationControlMechanism.agent_rep>` a Composition other than the
one to which the OptimizationControlMechanism belongs (and for which it is the `controller <Composition.controller>`).
In each `TRIAL <TimeScale.TRIAL>`, the `agent_rep <OptimizationControlMechanism.agent_rep>` is given the chance to
adapt, by adjusting its parameters in order to improve its prediction of the `net_outcome
<ControlMechanism.net_outcome>` for the Composition (or part of one) that is controlled by the
OptimizationControlMechanism (based on the `state <OptimizationControlMechanism_State>` and `net_outcome
<ControlMechanism.net_outcome>` of the prior trial).  The `agent_rep <OptimizationControlMechanism.agent_rep>` is
then used to predict the `net_outcome <ControlMechanism.net_outcome>` for `control_allocation
<ControlMechanism.control_allocation>` samples to find the one that yields the best predicted `net_outcome
<ControlMechanism.net_outcome>` of processing on the upcoming trial, based on the current or (expected)
`feature_values <OptimizationControlMechanism.feature_values>` for that trial.

.. _OptimizationControlMechanism_Model_Based:

*Model-Based Optimization*

This is achieved by assigning as the `agent_rep  <OptimizationControlMechanism.agent_rep>` the Composition to which the
OptimizationControlMechanism belongs (and for which it is the `controller <Composition.controller>`). On each `TRIAL
<TimeScale.TRIAL>`, that Composition itself is used to simulate processing on the upcoming trial, based on the current
or (expected) `feature_values <OptimizationControlMechanism.feature_values>` for that trial, in order to find the
<ControlMechanism.control_allocation>` that yields the best net_outcome <ControlMechanism.net_outcome>` for that trial.

.. _OptimizationControlMechanism_Creation:

Creating an OptimizationControlMechanism
----------------------------------------

An OptimizationControlMechanism is created in the same was as any `ControlMechanism <ControlMechanism>`.
The following arguments of its constructor are specific to the OptimizationControlMechanism:

* **features** -- takes the place of the standard **input_ports** argument in the constructor for a Mechanism`,
  and specifies the values used by the OptimizationControlMechanism, together with a `control_allocation
  <ControlMechanism.control_allocation>`, to calculate a `net_outcome <ControlMechanism.net_outcome>`.  For
  `model-based optimzation <OptimizationControlMechanism_Model_Based>` these are also used as the inputs to the
  Composition (i.e., `agent_rep <OptimizationControlMechanism.agent_rep>`) when it's `evaluate <Composition.evaluate>`
  method is called (see `OptimizationControlMechanism_Features` below).  Features can be specified using any of the
  following, singly or combined in a list:

  * *InputPort specification* -- this can be any form of `InputPort specification <InputPort_Specification>`
    that resolves to an OutputPort from which the InputPort receives a Projection;  the `value
    <OutputPort.value>` of that OutputPort is used as the feature. Each of these InputPorts is marked as
    `internal_only <InputPorts.internal_only>` = `True`.

  Features can also be added to an existing OptimizationControlMechanism using its `add_features` method.  If the
  **features** argument is not specified, then the `input <Composition.input_values>` to the `Composition` on the last
  trial of its execution is used to predict the `net_outcome <ControlMechanism.net_outcome>` for the upcoming trial.

.. _OptimizationControlMechanism_Feature_Function:

* **feature_function** -- specifies `function <InputPort>` of the InputPort created for each item listed in
  **features**.  By default, this is the identity function, that assigns the current value of the feature
  to the OptimizationControlMechanism's `feature_values <OptimizationControlMechanism.feature_values>` attribute.
  However, other functions can be assigned, for example to maintain a record of past values, or integrate them over
  trials.
..
* **agent_rep** -- specifies the `Composition` used by the OptimizationControlMechanism's `evaluation_function
  <OptimizationControlMechanism.evaluation_function>` to calculate the predicted `net_outcome
  <ControlMechanism.net_outcome>` for a given `state <OptimizationControlMechanism_State>` (see `below
  <OptimizationControlMechanism_Agent_Rep>` for additional details). If it is not specified, then the
  `Composition` to which the OptimizationControlMechanism is assigned becomes its `agent_rep
  <OptimizationControlMechanism.agent_rep>`, and the OptimizationControlMechanism is assigned as that Composition's
  `controller <Composition.controller>`, implementing fully `model-based <OptimizationControlMechanism_Model_Based>`
  optimization.  If that Composition already has a `controller <Composition.controller>` specified,
  the OptimizationControlMechanism is disabled. If another Composition is specified, it must conform to the
  specifications for an `agent_rep <OptimizationControlMechanism.agent_rep>` as described `below
  <OptimizationControlMechanism_Agent_Rep>`.

.. _OptimizationControlMechanism_Structure:

Structure
---------

In addition to the standard Components associated with a `ControlMechanism`, including a `Projection <Projection>`
to its *OUTCOME* InputPort from its `objective_mechanism <ControlMechanism.objective_mechanism>`, and a
`function <OptimizationControlMechanism.function>` used to carry out the optimization process, it has several
other constiuents, as described below.

.. _OptimizationControlMechanism_ObjectiveMechanism:

*ObjectiveMechanism*
^^^^^^^^^^^^^^^^^^^^

Like any `ControlMechanism`, an OptimizationControlMechanism may be assigned an `objective_mechanism
<ControlMechanism.objective_mechanism>` that is used to evaluate the outcome of processing for a given trial (see
`ControlMechanism_Objective_ObjectiveMechanism). This passes the result to the OptimizationControlMechanism, which it
places in its `outcome <OptimizationControlMechanism.outcome>` attribute.  This is used by its `compute_net_outcome
<ControlMechanism.compute_net_outcome>` function, together with the `costs <ControlMechanism.costs>` of its
`control_signals <ControlMechanism.control_signals>`, to compute the `net_outcome <ControlMechanism.net_outcome>` of
processing for a given `state <OptimizationControlMechanism_State>`, and that is returned by `evaluation` method of the
OptimizationControlMechanism's `agent_rep <OptimizationControlMechanism.agent_rep>`.

.. note::
    The `objective_mechanism <ControlMechanism.objective_mechanism>` is distinct from, and should not be
    confused with the `objective_function <OptimizationFunction.objective_function>` parameter of the
    OptimizationControlMechanism's `function <OptimizationControlMechanism.function>`.  The `objective_mechanism
    <ControlMechanism.objective_mechanism>` evaluates the `outcome <ControlMechanism.outcome>` of processing
    without taking into account the `costs <ControlMechanism.costs>` of the OptimizationControlMechanism's
    `control_signals <ControlMechanism.control_signals>`.  In contrast, its `evaluation_function
    <OptimizationControlMechanism.evaluation_function>`, which is assigned as the
    `objective_function` parameter of its `function <OptimizationControlMechanism.function>`, takes the `costs
    <ControlMechanism.costs>` of the OptimizationControlMechanism's `control_signals <ControlMechanism.control_signals>`
    into account when calculating the `net_outcome` that it returns as its result.

.. _OptimizationControlMechanism_Features:

*Features*
^^^^^^^^^^

In addition to its `primary InputPort <InputPort_Primary>` (which typically receives a projection from the
*OUTCOME* OutputPort of the `objective_mechanism <ControlMechanism.objective_mechanism>`,
an OptimizationControlMechanism also has an `InputPort` for each of its features. By default, these are the current
`input <Composition.input_values>` for the Composition to which the OptimizationControlMechanism belongs.  However,
different values can be specified, as can a `feature_function <OptimizationControlMechanism_Feature_Function>` that
transforms these.  For OptimizationControlMechanisms that implement `model-free
<OptimizationControlMechanism_Model_Free>` optimization, its `feature_values
<OptimizationControlMechanism.feature_values>` are used by its `evaluation_function
<OptimizationControlMechanism.evaluation_function>` to predict the `net_outcome <ControlMechanism.net_outcome>` for a
given `control_allocation <ControlMechanism.control_allocation>`.  For OptimizationControlMechanisms that implement
`model-based <OptimizationControlMechanism_Model_Based>` optimization, the `feature_values
<OptimizationCozntrolMechanism.feature_values>` are used as the Composition's `input <Composition.input_values>` when
it is executed to evaluate the `net_outcome <ControlMechanism.net_outcome>` for a given
`control_allocation<ControlMechanism.control_allocation>`.

Features can be of two types:

* *Input Features* -- these are values received as input by other Mechanisms in the `Composition`. They are
  specified as `shadowed inputs <InputPort_Shadow_Inputs>` in the **features** argument of the
  OptimizationControlMechanism's constructor (see `OptimizationControlMechanism_Creation`).  An InputPort is
  created on the OptimziationControlMechanism for each feature, that receives a `Projection` paralleling
  the input to be shadowed.
..
* *Output Features* -- these are the `value <OutputPort.value>` of an `OutputPort` of some other `Mechanism <Mechanism>`
  in the Composition.  These too are specified in the **features** argument of the OptimizationControlMechanism's
  constructor (see `OptimizationControlMechanism_Creation`), and each is assigned a `Projection` from the specified
  OutputPort(s) to the InputPort of the OptimizationControlMechanism for that feature.

The current `value <InputPort.value>` of the InputPorts for the features are listed in the `feature_values
<OptimizationControlMechanism.feature_values>` attribute.

.. _OptimizationControlMechanism_State:

*State*
^^^^^^^

The state of the Composition (or part of one) controlled by an OptimizationControlMechanism is defined by a combination
of `feature_values <OptimizationControlMechanism.feature_values>` (see `above <OptimizationControlMechanism_Features>`)
and a `control_allocation <ControlMechanism.control_allocation>`.

.. _OptimizationControlMechanism_Agent_Rep:

*Agent Representation*
^^^^^^^^^^^^^^^^^^^^^^

The defining feature of an OptimizationControlMechanism is its agent representation, specified in the **agent_rep**
argument of its constructor and assigned to its `agent_rep <OptimizationControlMechanism.agent_rep>` attribute.  This
designates a representation of the `Composition` (or parts of one) that the OptimizationControlMechanism controls, that
is used to evaluate sample `control_allocations <ControlMechanism.control_allocation>` in order to find the one that
optimizes the `net_outcome <ControlMechanism.net_outcome>`. The `agent_rep <OptimizationControlMechanism.agent_rep>`
is always itself a `Composition`, that can be either the same one that the OptimizationControlMechanism controls or
another one that is used to estimate the `net_outcome <ControlMechanism.net_outcome>` for that Composition (see `above
<OptimizationControlMechanism_Agent_Representation_Types>`).  The `evaluate <Composition.evaluate>` method of the
Composition is assigned as the `evaluation_function <OptimizationControlMechanism.evaluation_function>` of the
OptimizationControlMechanism.  If the `agent_rep <OptimizationControlMechanism.agent_rep>` is not the Composition for
which the OptimizationControlMechanism is the controller, then it must meet the following requirements:

    * Its `evaluate <Composition.evaluate>` method must accept as its first three arguments, in order,
      values that correspond in shape to  the `feature_values <OptimizationControlMechanism.feature_values>`,
      `control_allocation <ControlMechanism.control_allocation>` and `num_estimates
      <OptimizationControlMechanism.num_estimates>` attributes of the OptimizationControlMechanism, respectively.
    ..
    * If it has an `adapt <Composition.adapt>` method, that must accept as its first three arguments, in order,
      values that corresopnd to the shape of the `feature_values <OptimizationControlMechanism.feature_values>`,
      `control_allocation <ControlMechanism.control_allocation>` and `net_outcome
      <OptimizationControlMechanism.net_outcome>` attributes of the OptimizationControlMechanism. respectively.

.. _OptimizationControlMechanism_Function:

*Function*
^^^^^^^^^^

The `function <OptimizationControlMechanism.function>` of an OptimizationControlMechanism is used to find the
`control_allocation <ControlMechanism.control_allocation>` that optimizes the `net_outcome
<ControlMechanism.net_outcome>` for the current (or expected) `features <OptimizationControlMechanism_State>`.
It is generally an `OptimizationFunction`, which in turn has `objective_function
<OptimizationFunction.objective_function>`, `search_function <OptimizationFunction.search_function>`
and `search_termination_function <OptimizationFunction.search_termination_function>` methods, as well as a `search_space
<OptimizationFunction.search_space>` attribute.  The OptimizationControlMechanism's `evaluation_function
<OptimizationControlMechanism.evaluation_function>` is automatically assigned as the
OptimizationFunction's `objective_function <OptimizationFunction.objective_function>`, and is used to
evaluate each `control_allocation <ControlMechanism.control_allocation>` sampled from the `search_space
<OptimizationFunction.search_space>` by the `search_function `search_function <OptimizationFunction.search_function>`
until the `search_termination_function <OptimizationFunction.search_termination_function>` returns `True`.
A custom function can be assigned as the `function <OptimizationControlMechanism.function>` of an
OptimizationControlMechanism, however it must meet the following requirements:

.. _OptimizationControlMechanism_Custom_Funtion:

  - It must accept as its first argument and return as its result an array with the same shape as the
    OptimizationControlMechanism's `control_allocation <ControlMechanism.control_allocation>`.
  ..
  - It must implement a `reset` method that accepts **objective_function** as a keyword argument and
    implements an attribute with the same name.

    COMMENT:
    - it must implement a `reset` method that accepts as keyword arguments **objective_function**,
      **search_function**, **search_termination_function**, and **search_space**, and implement attributes
      with corresponding names.
    COMMENT

If **function** argument is not specified, the `GridSearch` `OptimiziationFunction` is assigned as the default,
which evaluates the `net_outcome <ControlMechanism.net_outcome>` using the OptimizationControlMechanism's
`control_allocation_search_space <OptimizationControlMechanism.control_allocation_search_spaces>` as its
`search_space <OptimizationFunction.search_space>`, and returns the `control_allocation
<ControlMechanism.control_allocation>` that yields the greatest `net_outcome <ControlMechanism.net_outcome>`,
thus implementing a computation of `EVC <OptimizationControlMechanism_EVC>`.

COMMENT:
.. _OptimizationControlMechanism_Search_Functions:

*Search Function, Search Space and Search Termination Function*

Subclasses of OptimizationControlMechanism may implement their own `search_function
<OptimizationControlMechanism.search_function>` and `search_termination_function
<OptimizationControlMechanism.search_termination_function>` methods, as well as a
`control_allocation_search_space <OptimizationControlMechanism.control_allocation_search_space>` attribute, that are
passed as parameters to the `OptimizationFunction` when it is constructed.  These can be specified in the
constructor for an `OptimizationFunction` assigned as the **function** argument in the
OptimizationControlMechanism's constructor, as long as they are compatible with the requirements of
the OptimizationFunction and OptimizationControlMechanism.  If they are not specified, then defaults specified
either by the OptimizationControlMechanism or the OptimizationFunction are used.
COMMENT

.. _OptimizationControlMechanism_Execution:

Execution
---------

When an OptimizationControlMechanism is executed, it carries out the following steps:

  * Calls `adapt` method of its `agent_rep <OptimizationControlMechanism.agent_rep>` to give that a chance to
    modify its parameters in order to better predict the `net_outcome <ControlMechanism.net_outcome>` for a given `state
    <OptimizationControlMechanism_State>`, based the state and `net_outcome <ControlMechanism.net_outcome>` of the
    previous trial.
  ..
  * Calls `function <OptimizationControlMechanism.function>` to find the `control_allocation
    <ControlMechanism.control_allocation>` that optimizes `net_outcome <ControlMechanism.net_outcome>`.  The
    way in which it searches for the best `control_allocation <ControlMechanism.control_allocation>` is determined by
    the type of `OptimzationFunction` assigned to `function <OptimizationControlMechanism.function>`, whereas the way
    that it evaluates each one is determined by the OptimizationControlMechanism's `evaluation_function
    <OptimizationControlMechanism.evalutation_function>`.  More specifically:

    * The `function <OptimizationControlMechanism.function>` selects a sample `control_allocation
      <ControlMechanism.control_allocation>` (using its `search_function <OptimizationFunction.search_function>`
      to select one from its `search_space <OptimizationFunction.search_space>`), and evaluates the predicted
      `net_outcome <ControlMechanism.net_outcome>` for that `control_allocation
      <ControlMechanism.control_allocation>` using the OptimizationControlMechanism's `evaluation_function`
      <OptimizationControlMechanism.evalutation_function>` and the current `feature_values
      <OptimizationControlMechanism.feature_values>`.
    ..
    * It continues to evaluate the `net_outcome <ControlMechanism.net_outcome>` for `control_allocation
      <ControlMechanism.control_allocation>` samples until its `search_termination_function
      <OptimizationFunction.search_termination_function>` returns `True`.
    ..
    * Finally, it implements the `control_allocation <ControlMechanism.control_allocation>` that yielded the optimal
      `net_outcome <ControlMechanism.net_outcome>`.  This is used by the OptimizationControlMechanism's `control_signals
      <ControlMechanism.control_signals>` to compute their `values <ControlSignal.value>` which, in turn, are used by
      their `ControlProjections <ControlProjection>` to modulate the parameters they control when the Composition is
      next executed.

COMMENT:
.. _OptimizationControlMechanism_Examples:

Examples
--------

The table below lists `model-free <ModelFreeOptimizationControlMechanism>` and `model-based
<ModelBasedOptimizationControlMechanism` subclasses of OptimizationControlMechanisms.

.. table:: **Model-Free and Model-Based OptimizationControlMechanisms**

   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |                         |     *Model-Free*     |                           *Model-Based*                                                         |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |**Functions**            |`LVOCControlMechanism`| LVOMControlMechanism | MDPControlMechanism |`EVCControlMechanism`| ParameterEstimationMechanism |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |**learning_function**    |     `BayesGLM`       |        `pymc`        |    `BeliefUpdate`   |       *None*        |           `pymc`             |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |**function** *(primary)* |`GradientOptimization`|     `GridSearch`     |       `Sample`      |    `GridSearch`     |           `Sample`           |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |       *search_function* |  *follow_gradient*   |   *traverse_grid*    | *sample_from_dist*  |   *traverse_grid*   |      *sample_from_dist*      |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |    *objective_function* |    *compute_EVC*     |  *evaluate*,   |  *evaluate*,  |  *evaluate*,  |    *evaluate*,         |
   |                         |                      |  *compute_EVC*       |  *compute_EVC*      |  *compute_EVC*      |    *compute_likelihood*      |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+
   |             *execution* | *iterate w/in trial* |  *once per trial*    | *iterate w/in trial*| *iterate w/in trial*|     *iterate w/in trial*     |
   +-------------------------+----------------------+----------------------+---------------------+---------------------+------------------------------+

The following models provide examples of implementing the OptimizationControlMechanisms in the table above:

`LVOCControlMechanism`\\:  `BustamanteStroopXORLVOCModel`
`EVCControlMechanism`\\:  `UmemotoTaskSwitchingEVCModel`
COMMENT



.. _OptimizationControlMechanism_Class_Reference:

Class Reference
---------------

"""
import copy
import numpy as np
import typecheck as tc

from collections.abc import Iterable

from psyneulink.core.components.component import DefaultsFlexibility
from psyneulink.core.components.functions.function import is_function_type, FunctionError
from psyneulink.core.components.functions.optimizationfunctions import \
    GridSearch, OBJECTIVE_FUNCTION, SEARCH_SPACE
from psyneulink.core.components.functions.combinationfunctions import LinearCombination
from psyneulink.core.components.functions.transferfunctions import CostFunctions
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.components.mechanisms.processing.objectivemechanism import \
    ObjectiveMechanism, ObjectiveMechanismError
from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import ControlMechanism
from psyneulink.core.components.shellclasses import Function
from psyneulink.core.components.ports.inputport import InputPort, _parse_shadow_inputs
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.components.ports.port import _parse_port_spec
from psyneulink.core.globals.context import Context, ContextFlags
from psyneulink.core.globals.defaults import defaultControlAllocation
from psyneulink.core.globals.keywords import \
    DEFAULT_VARIABLE, EID_FROZEN, FUNCTION, INTERNAL_ONLY, NAME, \
    OPTIMIZATION_CONTROL_MECHANISM, OBJECTIVE_MECHANISM, OUTCOME, PRODUCT, PARAMS, \
    CONTROL, AUTO_ASSIGN_MATRIX
from psyneulink.core.globals.utilities import convert_to_np_array
from psyneulink.core.globals.parameters import Parameter, ParameterAlias
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.context import handle_external_context
from psyneulink.core.globals.sampleiterator import SampleIterator, SampleSpec

from psyneulink.core import llvm as pnlvm

__all__ = [
    'OptimizationControlMechanism', 'OptimizationControlMechanismError',
    'AGENT_REP', 'FEATURES'
]

AGENT_REP = 'agent_rep'
FEATURES = 'features'


def _parse_feature_values_from_variable(variable):
    return convert_to_np_array(np.array(variable[1:]).tolist())


class OptimizationControlMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


def _control_allocation_search_space_getter(owning_component=None, context=None):
    search_space = owning_component.parameters.search_space._get(context)
    if not search_space:
        return [c.parameters.allocation_samples._get(context) for c in owning_component.control_signals]
    else:
        return search_space

class OptimizationControlMechanism(ControlMechanism):
    """OptimizationControlMechanism(         \
        objective_mechanism=None,            \
        monitor_for_control=None,            \
        objective_mechanism=None,            \
        origin_objective_mechanism=False     \
        terminal_objective_mechanism=False   \
        features=None,                       \
        feature_function=None,               \
        function=GridSearch,                       \
        agent_rep=None,                      \
        search_function=None,                \
        search_termination_function=None,    \
        search_space=None,                   \
        control_signals=None,                \
        modulation=MULTIPLICATIVE,           \
        combine_costs=np.sum,                \
        compute_reconfiguration_cost=None,   \
        compute_net_outcome=lambda x,y:x-y)

    Subclass of `ControlMechanism <ControlMechanism>` that adjusts its `ControlSignals <ControlSignal>` to optimize
    performance of the `Composition` to which it belongs.  See parent class for additional arguments.

    Arguments
    ---------

    features : Mechanism, OutputPort, Projection, dict, or list containing any of these
        specifies Components, the values of which are assigned to `feature_values
        <OptimizationControlMechanism.feature_values>` and used to predict `net_outcome <ControlMechanism.net_outcome>`.
        Any `InputPort specification <InputPort_Specification>` can be used that resolves to an `OutputPort` that
        projects to the InputPort.

    feature_function : Function or function : default None
        specifies the `function <InputPort.function>` for the `InputPort` assigned to each `feature
        <OptimizationControlMechanism_Features>`.

    agent_rep : None  : default Composition to which the OptimizationControlMechanism is assigned
        specifies the `Composition` used by the `evalution_function <OptimizationControlMechanism.evaluation_function>`
        to predict the `net_outcome <ControlMechanism.net_outcome>` for a given `state
        <OptimizationControlMechanism_State>`.  If a Composition is specified, it must be suitably configured
        (see `above <OptimizationControlMechanism_Agent_Rep>` for additional details). If it is not specified, the
        OptimizationControlMechanism is placed in `deferred_init` status until it is assigned as the `controller
        <Composition.controller>` of a Composition, at which that Composition is assigned as the `agent_rep
        <agent_rep <OptimizationControlMechanism.agent_rep`.

    search_function : function or method
        specifies the function assigned to `function <OptimizationControlMechanism.function>` as its
        `search_function <OptimizationFunction.search_function>` parameter, unless that is specified in a
        constructor for `function <OptimizationControlMechanism.function>`.  It must take as its arguments
        an array with the same shape as `control_allocation <ControlMechanism.control_allocation>` and an integer
        (indicating the iteration of the `optimization process <OptimizationFunction_Process>`), and return
        an array with the same shape as `control_allocation <ControlMechanism.control_allocation>`.

    search_termination_function : function or method
        specifies the function assigned to `function <OptimizationControlMechanism.function>` as its
        `search_termination_function <OptimizationFunction.search_termination_function>` parameter, unless that is
        specified in a constructor for `function <OptimizationControlMechanism.function>`.  It must take as its
        arguments an array with the same shape as `control_allocation <ControlMechanism.control_allocation>` and two
        integers (the first representing the `net_outcome <ControlMechanism.net_outcome>` for the current
        `control_allocation <ControlMechanism.control_allocation>`, and the second the current iteration of the
        `optimization process <OptimizationFunction_Process>`);  it must return `True` or `False`.

    search_space : iterable [list, tuple, ndarray, SampleSpec, or SampleIterator] | list, tuple, ndarray, SampleSpec, or SampleIterator
        specifies the `search_space <OptimizationFunction.search_space>` parameter for `function
        <OptimizationControlMechanism.function>`, unless that is specified in a constructor for `function
        <OptimizationControlMechanism.function>`.  An element at index i should correspond to an element at index i in
        `control_allocation <ControlMechanism.control_allocation>`. If
        `control_allocation <ControlMechanism.control_allocation>` contains only one element, then search_space can be
        specified as a single element without an enclosing iterable.

    function : OptimizationFunction, function or method
        specifies the function used to optimize the `control_allocation <ControlMechanism.control_allocation>`;
        must take as its sole argument an array with the same shape as `control_allocation
        <ControlMechanism.control_allocation>`, and return a similar array (see `Function
        <OptimizationControlMechanism_Function>` for additional details).

    Attributes
    ----------

    feature_values : 2d array
        the current value of each of the OptimizationControlMechanism's `features
        <OptimizationControlMechanism_Features>` (each of which is a 1d array).

    agent_rep : Composition
        determines the `Composition` used by the `evalution_function <OptimizationControlMechanism.evaluation_function>`
        to predict the `net_outcome <ControlMechanism.net_outcome>` for a given `state
        <OptimizationControlMechanism_State>` (see `above <OptimizationControlMechanism_Agent_Rep>`for additional
        details).

    function : OptimizationFunction, function or method
        takes current `control_allocation <ControlMechanism.control_allocation>` (as initializer),
        uses its `search_function <OptimizationFunction.search_function>` to select samples of `control_allocation
        <ControlMechanism.control_allocation>` from its `search_space <OptimizationFunction.search_space>`,
        evaluates these using its `evaluation_function <OptimizationControlMechanism.evaluation_function>`, and returns
        the one that yields the optimal `net_outcome <ControlMechanism.net_outcome>` (see `Function
        <OptimizationControlMechanism_Function>` for additional details).

    evaluation_function : function or method
        returns `net_outcome <ControlMechanism.net_outcome>` for a given `state <OptimizationControlMechanism_State>`
        (i.e., combination of `feature_values <OptimizationControlMechanism.feature_values>` and `control_allocation
        <ControlMechanism.control_allocation>`. It is assigned as the `objective_function
        <OptimizationFunction.objective_function>` parameter of `function
        <OptimizationControlMechanism.function>`, and calls the `evaluate` method of the OptimizationControlMechanism's
        `agent_rep <OptimizationControlMechanism.agent_rep>` with a `control_allocation
        <ControlMechanism.control_allocation>`, the OptimizationControlMechanism's `num_estimates
        <OptimizationControlMechanism.num_estimates>` attribute, and the current `feature_values
        <OptimizationControlMechanism.feature_values>`.

    COMMENT:
    search_function : function or method
        `search_function <OptimizationFunction.search_function>` assigned to `function
        <OptimizationControlMechanism.function>`; used to select samples of `control_allocation
        <ControlMechanism.control_allocation>` to evaluate by `evaluation_function
        <OptimizationControlMechanism.evaluation_function>`.

    search_termination_function : function or method
        `search_termination_function <OptimizationFunction.search_termination_function>` assigned to
        `function <OptimizationControlMechanism.function>`;  determines when to terminate the
        `optimization process <OptimizationFunction_Process>`.
    COMMENT

    control_allocation_search_space : list of SampleIterators
        `search_space <OptimizationFunction.search_space>` assigned by default to `function
        <OptimizationControlMechanism.function>`, that determines the samples of
        `control_allocation <ControlMechanism.control_allocation>` evaluated by the `evaluation_function
        <OptimizationControlMechanism.evaluation_function>`.  This is a proprety that, unless overridden,
        returns a list of the `SampleIterators <SampleIterator>` generated from the `allocation_sample
        <ControlSignal.allocation_sample>` specifications for each of the OptimizationControlMechanism's
        `control_signals <ControlMechanism.control_signals>`.

    saved_samples : list
        contains all values of `control_allocation <ControlMechanism.control_allocation>` sampled by `function
        <OptimizationControlMechanism.function>` if its `save_samples <OptimizationFunction.save_samples>` parameter
        is `True`;  otherwise list is empty.

    saved_values : list
        contains values of `net_outcome <ControlMechanism.net_outcome>` associated with all samples of
        `control_allocation <ControlMechanism.control_allocation>` evaluated by by `function
        <OptimizationControlMechanism.function>` if its `save_values <OptimizationFunction.save_samples>` parameter
        is `True`;  otherwise list is empty.

    search_statefulness : bool : True
        if set to False, an `OptimizationControlMechanism`\\ 's `evaluation_function` will never run simulations; the
        evaluations will simply execute in the original `execution context <_Execution_Contexts>`.

        if set to True, `simulations <OptimizationControlMechanism_Execution>` will be created normally for each
        `control allocation <control_allocation>`.

    """

    componentType = OPTIMIZATION_CONTROL_MECHANISM
    # initMethod = INIT_FULL_EXECUTE_METHOD
    # initMethod = INIT_EXECUTE_METHOD_ONLY

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to Type automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'DefaultControlMechanismCustomClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    # FIX: ADD OTHER Parameters() HERE??
    class Parameters(ControlMechanism.Parameters):
        """
            Attributes
            ----------

                agent_rep
                    see `agent_rep <OptimizationControlMechanism_Agent_Rep>`

                    :default value: None
                    :type:

                comp_execution_mode
                    see `comp_execution_mode <OptimizationControlMechanism.comp_execution_mode>`

                    :default value: `PYTHON`
                    :type: ``str``

                control_allocation_search_space
                    see `control_allocation_search_space <OptimizationControlMechanism.control_allocation_search_space>`

                    :default value: None
                    :type:

                feature_function
                    see `feature_function <OptimizationControlMechanism_Feature_Function>`

                    :default value: None
                    :type:

                function
                    see `function <OptimizationControlMechanism_Function>`

                    :default value: None
                    :type:

                input_ports
                    see `input_ports <OptimizationControlMechanism.input_ports>`

                    :default value: ["{name: OUTCOME, params: {internal_only: True}}"]
                    :type: ``list``
                    :read only: True

                num_estimates
                    see `num_estimates <OptimizationControlMechanism.num_estimates>`

                    :default value: None
                    :type:

                saved_samples
                    see `saved_samples <OptimizationControlMechanism.saved_samples>`

                    :default value: None
                    :type:

                saved_values
                    see `saved_values <OptimizationControlMechanism.saved_values>`

                    :default value: None
                    :type:

                search_function
                    see `search_function <OptimizationControlMechanism.search_function>`

                    :default value: None
                    :type:

                search_statefulness
                    see `search_statefulness <OptimizationControlMechanism.search_statefulness>`

                    :default value: True
                    :type: ``bool``

                search_termination_function
                    see `search_termination_function <OptimizationControlMechanism.search_termination_function>`

                    :default value: None
                    :type:
        """
        function = Parameter(GridSearch, stateful=False, loggable=False)
        feature_function = Parameter(None, reference=True, stateful=False, loggable=False)
        search_function = Parameter(None, stateful=False, loggable=False)
        search_space = Parameter(None, read_only=True)
        search_termination_function = Parameter(None, stateful=False, loggable=False)
        comp_execution_mode = Parameter('Python', stateful=False, loggable=False, pnl_internal=True)
        search_statefulness = Parameter(True, stateful=False, loggable=False)

        agent_rep = Parameter(None, stateful=False, loggable=False, pnl_internal=True, structural=True)

        feature_values = Parameter(_parse_feature_values_from_variable([defaultControlAllocation]), user=False, pnl_internal=True)

        input_ports = Parameter(
            [{NAME: OUTCOME, PARAMS: {INTERNAL_ONLY: True}}],
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
            parse_spec=True,
            aliases='features',
            constructor_argument='features'
        )
        num_estimates = None
        # search_space = None
        control_allocation_search_space = Parameter(None, read_only=True, getter=_control_allocation_search_space_getter)

        saved_samples = None
        saved_values = None

    @handle_external_context()
    @tc.typecheck
    def __init__(self,
                 agent_rep=None,
                 function=None,
                 features: tc.optional(tc.optional(tc.any(Iterable, Mechanism, OutputPort, InputPort))) = None,
                 feature_function: tc.optional(tc.optional(tc.any(is_function_type))) = None,
                 num_estimates = None,
                 search_function: tc.optional(tc.optional(tc.any(is_function_type))) = None,
                 search_termination_function: tc.optional(tc.optional(tc.any(is_function_type))) = None,
                 search_statefulness=None,
                 context=None,
                 **kwargs):
        """Implement OptimizationControlMechanism"""

        function = function or GridSearch

        # If agent_rep hasn't been specified, put into deferred init
        if agent_rep is None:
            if context.source==ContextFlags.COMMAND_LINE:
                # Temporarily name InputPort
                self._assign_deferred_init_name(self.__class__.__name__)
                # Store args for deferred initialization
                self._store_deferred_init_args(**locals())

                # Flag for deferred initialization
                self.initialization_status = ContextFlags.DEFERRED_INIT
                return
            # If constructor is called internally (i.e., for controller of Composition),
            # agent_rep needs to be specified
            else:
                assert False, f"PROGRAM ERROR: 'agent_rep' arg should have been specified " \
                              f"in internal call to constructor for {self.name}."

        super().__init__(
            function=function,
            input_ports=features,
            features=features,
            feature_function=feature_function,
            num_estimates=num_estimates,
            search_statefulness=search_statefulness,
            search_function=search_function,
            search_termination_function=search_termination_function,
            agent_rep=agent_rep,
            **kwargs
        )

    def _validate_params(self, request_set, target_set=None, context=None):
        """Insure that specification of ObjectiveMechanism has projections to it"""

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        from psyneulink.core.compositions.composition import Composition
        if request_set[AGENT_REP] is None:
            raise OptimizationControlMechanismError(f"The {repr(AGENT_REP)} arg of an {self.__class__.__name__} must "
                                                    f"be specified and be a {Composition.__name__}")

        elif not (isinstance(request_set[AGENT_REP], Composition)
                  or (isinstance(request_set[AGENT_REP], type) and issubclass(request_set[AGENT_REP], Composition))):
            raise OptimizationControlMechanismError(f"The {repr(AGENT_REP)} arg of an {self.__class__.__name__} "
                                                    f"must be either a {Composition.__name__} or a sublcass of one")

    def _instantiate_input_ports(self, context=None):
        """Instantiate input_ports for Projections from features and objective_mechanism.

        Inserts InputPort specification for Projection from ObjectiveMechanism as first item in list of
        InputPort specifications generated in _parse_feature_specs from the **features** and
        **feature_function** arguments of the OptimizationControlMechanism constructor.
        """

        # Specify *OUTCOME* InputPort;  receives Projection from *OUTCOME* OutputPort of objective_mechanism
        outcome_input_port = {NAME:OUTCOME, PARAMS:{INTERNAL_ONLY:True}}

        # If any features were specified (assigned to self.input_ports in __init__):
        if self.input_ports:
            self.input_ports = _parse_shadow_inputs(self, self.input_ports)
            self.input_ports = self._parse_feature_specs(self.input_ports, self.feature_function)
            # Insert primary InputPort for outcome from ObjectiveMechanism;
            #     assumes this will be a single scalar value and must be named OUTCOME by convention of ControlSignal
            self.input_ports.insert(0, outcome_input_port),
        else:
            self.input_ports = [outcome_input_port]

        # Configure default_variable to comport with full set of input_ports
        self.defaults.variable, _ = self._handle_arg_input_ports(self.input_ports)

        super()._instantiate_input_ports(context=context)

        for i in range(1, len(self.input_ports)):
            port = self.input_ports[i]
            if len(port.path_afferents) > 1:
                raise OptimizationControlMechanismError(f"Invalid {type(input_port).__name__} on {self.name}. "
                                                        f"{port.name} should receive exactly one projection, "
                                                        f"but it receives {len(port.path_afferents)} projections.")

    def _instantiate_output_ports(self, context=None):
        """Assign CostFunctions.DEFAULTS as default for cost_option of ControlSignals.
        OptimizationControlMechanism requires use of at least one of the cost options
        """
        super()._instantiate_output_ports(context)

        for control_signal in self.control_signals:
            if control_signal.cost_options is None:
                control_signal.cost_options = CostFunctions.DEFAULTS
                control_signal._instantiate_cost_attributes(context)

    def _instantiate_control_signals(self, context):
        """Size control_allocation and assign modulatory_signals
        Set size of control_allocadtion equal to number of modulatory_signals.
        Assign each modulatory_signal sequentially to corresponding item of control_allocation.
        """
        from psyneulink.core.globals.keywords import OWNER_VALUE
        output_port_specs = list(enumerate(self.output_ports))
        for i, spec in output_port_specs:
            control_signal = self._instantiate_control_signal(spec, context=context)
            control_signal._variable_spec = (OWNER_VALUE, i)
            self.output_ports[i] = control_signal
        self.defaults.value = np.tile(control_signal.parameters.variable.default_value, (i + 1, 1))
        self.parameters.control_allocation._set(copy.deepcopy(self.defaults.value), context)

    def _instantiate_function(self, function, function_params=None, context=None):
        # this indicates a significant peculiarity of OCM, in that its function
        # corresponds to its value (control_allocation) rather than anything to
        # do with its variable. see _instantiate_attributes_after_function

        # Workaround this issue here, and explicitly allow the function's
        # default variable to be modified
        if isinstance(function, Function):
            function._variable_shape_flexibility = DefaultsFlexibility.FLEXIBLE

        super()._instantiate_function(function, function_params, context)

    def _instantiate_attributes_after_function(self, context=None):
        """Instantiate OptimizationControlMechanism's OptimizatonFunction attributes"""

        super()._instantiate_attributes_after_function(context=context)

        search_space = self.parameters.search_space._get(context)
        if type(search_space) == np.ndarray:
            search_space = search_space.tolist()
        if search_space:
            corrected_search_space = []
            try:
                if type(search_space) == SampleIterator:
                    corrected_search_space.append(search_space)
                elif type(search_space) == SampleSpec:
                    corrected_search_space.append(SampleIterator(search_space))
                else:
                    for i in self.parameters.search_space._get(context):
                        if not type(i) == SampleIterator:
                            corrected_search_space.append(SampleIterator(specification=i))
                            continue
                        corrected_search_space.append(i)
            except AssertionError:
                corrected_search_space = [SampleIterator(specification=search_space)]
            self.parameters.search_space._set(corrected_search_space, context)

        # Assign parameters to function (OptimizationFunction) that rely on OptimizationControlMechanism
        self.function.reset(**{
            DEFAULT_VARIABLE: self.parameters.control_allocation._get(context),
            OBJECTIVE_FUNCTION: self.evaluation_function,
            # SEARCH_FUNCTION: self.search_function,
            # SEARCH_TERMINATION_FUNCTION: self.search_termination_function,
            SEARCH_SPACE: self.parameters.control_allocation_search_space._get(context)
        })

        if isinstance(self.agent_rep, type):
            self.agent_rep = self.agent_rep()

        from psyneulink.core.compositions.compositionfunctionapproximator import CompositionFunctionApproximator
        if (isinstance(self.agent_rep, CompositionFunctionApproximator)):
            self._initialize_composition_function_approximator(context)

    def _update_input_ports(self, runtime_params=None, context=None):
        """Update value for each InputPort in self.input_ports:

        Call execute method for all (MappingProjection) Projections in Port.path_afferents
        Aggregate results (using InputPort execute method)
        Update InputPort.value
        """
        # "Outcome"
        outcome_input_port = self.input_port
        outcome_input_port._update(params=runtime_params, context=context)
        port_values = [np.atleast_2d(outcome_input_port.parameters.value._get(context))]
        # MODIFIED 5/8/20 OLD:
        # FIX 5/8/20 [JDC]: THIS DOESN'T CALL SUPER, SO NOT IDEAL HOWEVER, REVISION BELOW CRASHES... NEEDS TO BE FIXED
        for i in range(1, len(self.input_ports)):
            port = self.input_ports[i]
            port._update(params=runtime_params, context=context)
            port_values.append(port.parameters.value._get(context))
        return convert_to_np_array(port_values)
        # # MODIFIED 5/8/20 NEW:
        # input_port_values = super()._update_input_ports(runtime_params, context)
        # port_values.append(input_port_values)
        # return np.array(port_values)
        # MODIFIED 5/8/20 END

    def _execute(self, variable=None, context=None, runtime_params=None):
        """Find control_allocation that optimizes result of `agent_rep.evaluate`  ."""

        if self.is_initializing:
            return [defaultControlAllocation]

        # # FIX: THESE NEED TO BE FOR THE PREVIOUS TRIAL;  ARE THEY FOR FUNCTION_APPROXIMATOR?
        self.parameters.feature_values._set(_parse_feature_values_from_variable(variable), context)

        # Assign default control_allocation if it is not yet specified (presumably first trial)
        control_allocation = self.parameters.control_allocation._get(context)
        if control_allocation is None:
            control_allocation = [c.defaults.variable for c in self.control_signals]
            self.parameters.control_allocation._set(control_allocation, context=None)

        # Give the agent_rep a chance to adapt based on last trial's feature_values and control_allocation
        if hasattr(self.agent_rep, "adapt"):
            # KAM 4/11/19 switched from a try/except to hasattr because in the case where we don't
            # have an adapt method, we also don't need to call the net_outcome getter
            net_outcome = self.parameters.net_outcome._get(context)

            self.agent_rep.adapt(_parse_feature_values_from_variable(variable),
                                 control_allocation,
                                 net_outcome,
                                 context=context)

        # freeze the values of current context, because they can be changed in between simulations,
        # and the simulations must start from the exact spot
        self.agent_rep._initialize_from_context(self._get_frozen_context(context),
                                                base_context=context,
                                                override=True)

        # Get control_allocation that optimizes net_outcome using OptimizationControlMechanism's function
        # IMPLEMENTATION NOTE: skip ControlMechanism._execute since it is a stub method that returns input_values
        optimal_control_allocation, optimal_net_outcome, saved_samples, saved_values = \
                                                super(ControlMechanism,self)._execute(variable=control_allocation,
                                                                                      context=context,
                                                                                      runtime_params=runtime_params,
                                                                                      )

        # clean up frozen values after execution
        self.agent_rep._delete_contexts(self._get_frozen_context(context))

        optimal_control_allocation = np.array(optimal_control_allocation).reshape((len(self.defaults.value), 1))
        if self.function.save_samples:
            self.saved_samples = saved_samples
        if self.function.save_values:
            self.saved_values = saved_values

        # Return optimal control_allocation
        return optimal_control_allocation

    def _get_frozen_context(self, context=None):
        return Context(execution_id=f'{context.execution_id}{EID_FROZEN}')

    def _set_up_simulation(self, base_context=Context(execution_id=None), control_allocation=None):
        sim_context = copy.copy(base_context)
        sim_context.execution_id = self.get_next_sim_id(base_context)

        if control_allocation is not None:
            sim_context.execution_id += f'-{control_allocation}'

        try:
            self.parameters.simulation_ids._get(base_context).append(sim_context.execution_id)
        except AttributeError:
            self.parameters.simulation_ids._set([sim_context.execution_id], base_context)

        self.agent_rep._initialize_from_context(sim_context, self._get_frozen_context(base_context), override=False)

        return sim_context

    def _tear_down_simulation(self, sim_context=None):
        if not self.agent_rep.parameters.retain_old_simulation_data._get():
            self.agent_rep._delete_contexts(sim_context, check_simulation_storage=True)

    def evaluation_function(self, control_allocation, context=None, return_results=False):
        """Compute `net_outcome <ControlMechanism.net_outcome>` for current set of `feature_values
        <OptimizationControlMechanism.feature_values>` and a specified `control_allocation
        <ControlMechanism.control_allocation>`.

        Assigned as the `objective_function <OptimizationFunction.objective_function>` for the
        OptimizationControlMechanism's `function <OptimizationControlMechanism.function>`.

        Calls `agent_rep <OptimizationControlMechanism.agent_rep>`\\'s `evalute` method.

        Returns a scalar that is the predicted `net_outcome <ControlMechanism.net_outcome>`
        for the current `feature_values <OptimizationControlMechanism.feature_values>`
        and specified `control_allocation <ControlMechanism.control_allocation>`.

        """
        # agent_rep is a Composition (since runs_simuluations = True)
        if self.agent_rep.runs_simulations:
            # KDM 5/20/19: crudely using default here because it is a stateless parameter
            # and there is a bug in setting parameter values on init, see TODO note above
            # call to self._instantiate_defaults around component.py:1115
            if self.defaults.search_statefulness:
                new_context = self._set_up_simulation(context, control_allocation)
            else:
                new_context = context

            old_composition = context.composition
            context.composition = self.agent_rep

            # We shouldn't get this far if execution mode is not Python
            exec_mode = self.parameters.comp_execution_mode._get(context)
            assert exec_mode == "Python"
            result = self.agent_rep.evaluate(self.parameters.feature_values._get(context),
                                             control_allocation,
                                             self.parameters.num_estimates._get(context),
                                             base_context=context,
                                             context=new_context,
                                             execution_mode=exec_mode,
                                             return_results=return_results)
            context.composition = old_composition

            if self.defaults.search_statefulness:
                self._tear_down_simulation(new_context)

            # If results of the simulation shoudld be returned then, do so. Agent Rep Evaluate will
            # return a tuple in this case where the first element is the outcome as usual and the
            # results of composision run are the second element.
            if return_results:
                return result[0], result[1]
            else:
                return result

        # agent_rep is a CompositionFunctionApproximator (since runs_simuluations = False)
        else:
            result = self.agent_rep.evaluate(self.parameters.feature_values._get(context),
                                             control_allocation,
                                             self.parameters.num_estimates._get(context),
                                             context=context
            )

        return result

    def _get_evaluate_input_struct_type(self, ctx):
        # We construct input from optimization function input
        return ctx.get_input_struct_type(self.function)

    def _get_evaluate_output_struct_type(self, ctx):
        # Returns a scalar that is the predicted net_outcome
        return ctx.float_ty

    def _get_evaluate_alloc_struct_type(self, ctx):
        return pnlvm.ir.ArrayType(ctx.float_ty,
                                  len(self.parameters.control_allocation_search_space.get()))

    def _gen_llvm_net_outcome_function(self, *, ctx, tags=frozenset()):
        assert "net_outcome" in tags
        args = [ctx.get_param_struct_type(self).as_pointer(),
                ctx.get_state_struct_type(self).as_pointer(),
                self._get_evaluate_alloc_struct_type(ctx).as_pointer(),
                ctx.float_ty.as_pointer(),
                ctx.float_ty.as_pointer()]

        builder = ctx.create_llvm_function(args, self, str(self) + "_net_outcome")
        llvm_func = builder.function
        for p in llvm_func.args:
            p.attributes.add('nonnull')
        params, state, allocation_sample, objective_ptr, arg_out = llvm_func.args

        op_params = pnlvm.helpers.get_param_ptr(builder, self, params,
                                                "output_ports")
        op_states = pnlvm.helpers.get_state_ptr(builder, self, state,
                                                "output_ports", None)

        # calculate cost function
        total_cost = builder.alloca(ctx.float_ty)
        builder.store(ctx.float_ty(-0.0), total_cost)
        for i, op in enumerate(self.output_ports):
            op_i_params = builder.gep(op_params, [ctx.int32_ty(0),
                                                  ctx.int32_ty(i)])
            op_i_state = builder.gep(op_states, [ctx.int32_ty(0),
                                                 ctx.int32_ty(i)])

            op_f = ctx.import_llvm_function(op, tags=frozenset({"costs"}))

            op_in = builder.alloca(op_f.args[2].type.pointee)

            # copy allocation_sample, the input is 1-element array in a struct
            data_in = builder.gep(allocation_sample, [ctx.int32_ty(0),
                                                      ctx.int32_ty(i)])
            data_out = builder.gep(op_in, [ctx.int32_ty(0), ctx.int32_ty(0),
                                           ctx.int32_ty(0)])
            builder.store(builder.load(data_in), data_out)

            # Invoke cost function
            cost = builder.call(op_f, [op_i_params, op_i_state, op_in])

            # simplified version of combination fmax(cost, 0)
            ltz = builder.fcmp_ordered("<", cost, cost.type(0))
            cost = builder.select(ltz, cost.type(0), cost)

            # combine is not a PNL function
            assert self.combine_costs is np.sum
            val = builder.load(total_cost)
            val = builder.fadd(val, cost)
            builder.store(val, total_cost)

        # compute net outcome
        objective = builder.load(objective_ptr)
        net_outcome = builder.fsub(objective, builder.load(total_cost))
        builder.store(net_outcome, arg_out)

        builder.ret_void()
        return llvm_func

    def _gen_llvm_evaluate_function(self, *, ctx:pnlvm.LLVMBuilderContext,
                                             tags=frozenset()):
        assert "evaluate" in tags
        args = [ctx.get_param_struct_type(self.agent_rep).as_pointer(),
                ctx.get_state_struct_type(self.agent_rep).as_pointer(),
                self._get_evaluate_alloc_struct_type(ctx).as_pointer(),
                self._get_evaluate_output_struct_type(ctx).as_pointer(),
                self._get_evaluate_input_struct_type(ctx).as_pointer(),
                ctx.get_data_struct_type(self.agent_rep).as_pointer()]

        builder = ctx.create_llvm_function(args, self, str(self) + "_evaluate")
        llvm_func = builder.function
        for p in llvm_func.args:
            p.attributes.add('nonnull')

        comp_params, base_comp_state, allocation_sample, arg_out, arg_in, base_comp_data = llvm_func.args

        # Create a simulation copy of composition state
        comp_state = builder.alloca(base_comp_state.type.pointee, name="state_copy")
        builder.store(builder.load(base_comp_state), comp_state)

        # Create a simulation copy of composition data
        comp_data = builder.alloca(base_comp_data.type.pointee, name="data_copy")
        builder.store(builder.load(base_comp_data), comp_data)

        # Evaluate is called on composition controller
        assert self.composition.controller is self
        assert self.composition is self.agent_rep
        nodes_states = pnlvm.helpers.get_state_ptr(builder, self.composition,
                                                   comp_state, "nodes", None)
        nodes_params = pnlvm.helpers.get_param_ptr(builder, self.composition,
                                                   comp_params, "nodes")

        controller_idx = self.composition._get_node_index(self)
        controller_state = builder.gep(nodes_states, [ctx.int32_ty(0),
                                                      ctx.int32_ty(controller_idx)])
        controller_params = builder.gep(nodes_params, [ctx.int32_ty(0),
                                                       ctx.int32_ty(controller_idx)])

        # Get simulation function
        sim_f = ctx.import_llvm_function(self.agent_rep,
                                         tags=frozenset({"run", "simulation"}))

        # Apply allocation sample to simulation data
        assert len(self.output_ports) == len(allocation_sample.type.pointee)
        idx = self.agent_rep._get_node_index(self)
        ocm_out = builder.gep(comp_data, [ctx.int32_ty(0), ctx.int32_ty(0),
                                          ctx.int32_ty(idx)])
        for i, _ in enumerate(self.output_ports):
            idx = ctx.int32_ty(i)
            sample_ptr = builder.gep(allocation_sample, [ctx.int32_ty(0), idx])
            sample_dst = builder.gep(ocm_out, [ctx.int32_ty(0), idx, ctx.int32_ty(0)])
            if sample_ptr.type != sample_dst.type:
                assert len(sample_dst.type.pointee) == 1
                sample_dst = builder.gep(sample_dst, [ctx.int32_ty(0),
                                                      ctx.int32_ty(0)])
            builder.store(builder.load(sample_ptr), sample_dst)

        # Construct input
        comp_input = builder.alloca(sim_f.args[3].type.pointee, name="sim_input")

        for src_idx, ip in enumerate(self.input_ports):
            if ip.shadow_inputs is None:
                continue
            cim_in_port = self.agent_rep.input_CIM_ports[ip.shadow_inputs][0]
            dst_idx = self.agent_rep.input_CIM.input_ports.index(cim_in_port)

            src = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(src_idx)])
            # Destination is a struct of 2d arrays
            dst = builder.gep(comp_input, [ctx.int32_ty(0),
                                           ctx.int32_ty(dst_idx),
                                           ctx.int32_ty(0)])
            builder.store(builder.load(src), dst)


        # Determine simulation counts
        num_estimates_ptr = pnlvm.helpers.get_param_ptr(builder, self,
                                                        controller_params,
                                                        "num_estimates")

        num_estimates = builder.load(num_estimates_ptr, "num_estimates")

        # if num_estimates is 0, run 1 trial
        param_is_zero = builder.icmp_unsigned("==", num_estimates,
                                                    ctx.int32_ty(0))
        num_sims = builder.select(param_is_zero, ctx.int32_ty(1),
                                  num_estimates, "corrected_estimates")

        num_runs = builder.alloca(ctx.int32_ty, name="num_runs")
        builder.store(num_sims, num_runs)

        # We only provide one input
        num_inputs = builder.alloca(ctx.int32_ty, name="num_inputs")
        builder.store(num_inputs.type.pointee(1), num_inputs)

        # Simulations don't store output
        comp_output = sim_f.args[4].type(None)
        builder.call(sim_f, [comp_state, comp_params, comp_data, comp_input,
                             comp_output, num_runs, num_inputs])

        # Extract objective mechanism value
        idx = self.agent_rep._get_node_index(self.objective_mechanism)
        # Mechanisms' results are stored in the first substructure
        objective_os_ptr = builder.gep(comp_data, [ctx.int32_ty(0),
                                                   ctx.int32_ty(0),
                                                   ctx.int32_ty(idx)])
        # Objective mech output shape should be 1 single element 2d array
        objective_val_ptr = builder.gep(objective_os_ptr,
                                        [ctx.int32_ty(0), ctx.int32_ty(0),
                                         ctx.int32_ty(0)], "obj_val_ptr")

        net_outcome_f = ctx.import_llvm_function(self, tags=tags.union({"net_outcome"}))
        builder.call(net_outcome_f, [controller_params, controller_state,
                                     allocation_sample, objective_val_ptr,
                                     arg_out])

        builder.ret_void()

        return llvm_func

    def _gen_llvm_function(self, *, ctx:pnlvm.LLVMBuilderContext, tags:frozenset):
        if "net_outcome" in tags:
            return self._gen_llvm_net_outcome_function(ctx=ctx, tags=tags)
        if "evaluate" in tags:
            return self._gen_llvm_evaluate_function(ctx=ctx, tags=tags)

        is_comp = not isinstance(self.agent_rep, Function)
        if is_comp:
            extra_args = [ctx.get_param_struct_type(self.agent_rep).as_pointer(),
                          ctx.get_state_struct_type(self.agent_rep).as_pointer(),
                          ctx.get_data_struct_type(self.agent_rep).as_pointer()]
        else:
            extra_args = []

        f = super()._gen_llvm_function(ctx=ctx, extra_args=extra_args, tags=tags)
        if is_comp:
            for a in f.args[-len(extra_args):]:
                a.attributes.add('nonnull')

        return f

    def _gen_llvm_invoke_function(self, ctx, builder, function, params, context, variable, *, tags:frozenset):
        fun = ctx.import_llvm_function(function)
        fun_out = builder.alloca(fun.args[3].type.pointee)

        args = [params, context, variable, fun_out]
        # If we're calling compiled version of Composition.evaluate,
        # we need to pass extra arguments
        if len(fun.args) > 4:
            args += builder.function.args[-3:]
        builder.call(fun, args)

        return fun_out, builder

    def _gen_llvm_output_port_parse_variable(self, ctx, builder, params, context, value, port):
        i = self.output_ports.index(port)
        # Allocate the only member of the port input struct
        oport_input = builder.alloca(ctx.get_input_struct_type(port).elements[0])
        # FIXME: workaround controller signals occasionally being 2d
        dest_ptr = pnlvm.helpers.unwrap_2d_array(builder, oport_input)
        dest_ptr = builder.gep(dest_ptr, [ctx.int32_ty(0), ctx.int32_ty(0)])
        val_ptr = builder.gep(value, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(i)])
        builder.store(builder.load(val_ptr), dest_ptr)
        return oport_input

    # @property
    # def feature_values(self):
    #     if hasattr(self.agent_rep, 'model_based_optimizer') and self.agent_rep.model_based_optimizer is self:
    #         return self.agent_rep._get_predicted_input()
    #     else:
    #         return np.array(np.array(self.variable[1:]).tolist())

    # FIX: THE FOLLOWING SHOULD BE MERGED WITH HANDLING OF PredictionMechanisms FOR ORIG MODEL-BASED APPROACH;
    # FIX: SHOULD BE GENERALIZED AS SOMETHING LIKE update_feature_values

    @tc.typecheck
    @handle_external_context()
    def add_features(self, features, context=None):
        """Add InputPorts and Projections to OptimizationControlMechanism for features used to
        predict `net_outcome <ControlMechanism.net_outcome>`

        **features** argument can use any of the forms of specification allowed for InputPort(s)
        """

        if features:
            features = self._parse_feature_specs(features=features,
                                                 context=context)
        self.add_ports(InputPort, features)

    @tc.typecheck
    def _parse_feature_specs(self, input_ports, feature_function, context=None):
        """Parse entries of features into InputPort spec dictionaries
        Set INTERNAL_ONLY entry of params dict of InputPort spec dictionary to True
            (so that inputs to Composition are not required if the specified state is on an INPUT Mechanism)
        Assign functions specified in **feature_function** to InputPorts for all features
        Return list of InputPort specification dictionaries
        """

        parsed_features = []

        if not isinstance(input_ports, list):
            input_ports = [input_ports]

        for spec in input_ports:
            spec = _parse_port_spec(owner=self, port_type=InputPort, port_spec=spec)    # returns InputPort dict
            spec[PARAMS][INTERNAL_ONLY] = True
            if feature_function:
                if isinstance(feature_function, Function):
                    feat_fct = copy.deepcopy(feature_function)
                else:
                    feat_fct = feature_function
                spec.update({FUNCTION: feat_fct})
            spec = [spec]   # so that extend works below

            parsed_features.extend(spec)

        return parsed_features

    @property
    def _model_spec_parameter_blacklist(self):
        # default_variable is hidden in constructor arguments,
        # and anyway assigning it is problematic because it is modified
        # several times when creating input ports, and assigning function that
        # fits the control allocation
        return super()._model_spec_parameter_blacklist.union({
            'variable',
        })

    # ******************************************************************************************************************
    # FIX:  THE FOLLOWING IS SPECIFIC TO CompositionFunctionApproximator AS agent_rep
    # ******************************************************************************************************************

    def _initialize_composition_function_approximator(self, context):
        """Initialize CompositionFunctionApproximator"""

        # CompositionFunctionApproximator needs to have access to control_signals to:
        # - to construct control_allocation_search_space from their allocation_samples attributes
        # - compute their values and costs for samples of control_allocations from control_allocation_search_space
        self.agent_rep.initialize(features_array=np.array(self.defaults.variable[1:]),
                                  control_signals = self.control_signals,
                                  context=context)



from psyneulink import *

optimal_color_control = 1
optimal_motion_control = .5

color_input = ProcessingMechanism(name='Color',
                                  function=Linear(slope=optimal_color_control))
motion_input = ProcessingMechanism(name='Motion',
                                  function=Linear(slope=optimal_motion_control))
decision = DDM(name='Decision',
               function=DriftDiffusionAnalytical(
                       starting_point=0,
                       noise=0.5,
                       t0=0.2,
                       threshold=0.45),
               output_ports=[PROBABILITY_UPPER_THRESHOLD, RESPONSE_TIME],
               )

c = Composition(name='ColorMotion Task')
c.add_linear_processing_pathway([color_input, decision])
c.add_linear_processing_pathway([motion_input, decision])

# c.show_graph()
c.show_graph(show_node_structure=ALL)

stimuli = {color_input: [0,1,0],
              motion_input: [1,0,1]}

# c.run(inputs=stimuli)
# print (c.results)
print (c.json_summary())





import numpy as np
import psyneulink as pnl
import pytest

class TestComponent:

    def test_detection_of_legal_arg_in_kwargs(self):
        assert isinstance(pnl.ProcessingMechanism().reset_stateful_function_when, pnl.Never)
        assert isinstance(pnl.ProcessingMechanism(reset_stateful_function_when=pnl.AtTrialStart()).reset_stateful_function_when,
                          pnl.AtTrialStart)

    def test_detection_of_illegal_arg_in_kwargs(self):
        with pytest.raises(pnl.ComponentError) as error_text:
            pnl.ProcessingMechanism(flim_flam=1)
        assert "Unrecognized argument in constructor for ProcessingMechanism-0 (type: ProcessingMechanism): 'flim_flam'"

    def test_detection_of_illegal_args_in_kwargs(self):
        with pytest.raises(pnl.ComponentError) as error_text:
            pnl.ProcessingMechanism(name='MY_MECH', flim_flam=1, grumblabble=2)
        assert "Unrecognized arguments in constructor for MY_MECH (type: ProcessingMechanism): 'flim_flam, grumblabble'"

    def test_component_execution_counts_for_standalone_mechanism(self):
        """Note: input_port should not update execution count, since it has no afferents"""

        T = pnl.TransferMechanism()

        T.execute()
        assert T.execution_count == 1
        assert T.input_port.execution_count == 0

        # skipped (0 executions) because we bypass execute when no afferents, and
        # function._is_identity is satisfied (here, Linear function with slope 0 and intercept 1)
        # This holds true for each below
        assert T.parameter_ports[pnl.SLOPE].execution_count == 0
        assert T.output_port.execution_count == 0

        T.execute()
        assert T.execution_count == 2
        assert T.input_port.execution_count == 0
        assert T.parameter_ports[pnl.SLOPE].execution_count == 0
        assert T.output_port.execution_count == 0

        T.execute()
        assert T.execution_count == 3
        assert T.input_port.execution_count == 0
        assert T.parameter_ports[pnl.SLOPE].execution_count == 0
        assert T.output_port.execution_count == 0

    def test_component_execution_counts_for_mechanisms_in_composition(self):

        T1 = pnl.TransferMechanism()
        T2 = pnl.TransferMechanism()
        c = pnl.Composition()
        c.add_node(T1)
        c.add_node(T2)
        c.add_projection(sender=T1, receiver=T2)

        input_dict = {T1:[[0]]}

        c.run(input_dict)
        assert T2.execution_count == 1
        assert T2.input_port.execution_count == 1
        assert T2.parameter_ports[pnl.SLOPE].execution_count == 0
        assert T2.output_port.execution_count == 0

        c.run(input_dict)
        assert T2.execution_count == 2
        assert T2.input_port.execution_count == 2
        assert T2.parameter_ports[pnl.SLOPE].execution_count == 0
        assert T2.output_port.execution_count == 0

        c.run(input_dict)
        assert T2.execution_count == 3
        assert T2.input_port.execution_count == 3
        assert T2.parameter_ports[pnl.SLOPE].execution_count == 0
        assert T2.output_port.execution_count == 0

    def test__set_all_parameter_properties_recursively(self):
        A = pnl.ProcessingMechanism(name='A')
        A._set_all_parameter_properties_recursively(history_max_length=0)

        for c in A._dependent_components:
            for param in c.parameters:
                assert param.history_max_length == 0

    @pytest.mark.parametrize(
        'component_type', [
            pnl.ProcessingMechanism,
            pnl.TransferMechanism,
            pnl.Linear,
            pnl.DDM
        ]
    )
    def test_execute_manual_context(self, component_type):
        c = component_type()
        default_result = c.execute(5)

        assert pnl.safe_equals(c.execute(5, context='new'), default_result)



# import psyneulink as pnl
# import numpy as np
# import matplotlib.pyplot as plt
# import pandas
#
#
# def test_lauras_cohen_1990_model(red_color, green_color, red_word, green_word, CN, WR, n_trials):
#     #  INPUT UNITS
#
#     #  colors: ('red', 'green'), words: ('RED','GREEN')
#     colors_input_layer = pnl.TransferMechanism(size=2,
#                                                function=pnl.Linear,
#                                                name='COLORS_INPUT')
#
#     words_input_layer = pnl.TransferMechanism(size=2,
#                                               function=pnl.Linear,
#                                               name='WORDS_INPUT')
#
#     #   Task layer, tasks: ('name the color', 'read the word')
#     task_layer = pnl.TransferMechanism(size=2,
#                                        function=pnl.Linear,
#                                        name='TASK')
#
#     #   HIDDEN LAYER UNITS
#
#     #   colors_hidden: ('red','green')
#     #   Logistic activation function, Gain = 1.0, Bias = -4.0 (in PNL bias is subtracted so enter +4.0 to get negative bias)
#     #   randomly distributed noise to the net input
#     #   time averaging = integration_rate = 0.1
#     unit_noise = 0.005
#     # colors_hidden_layer = pnl.TransferMechanism(size=2,
#     #                                             function=pnl.Logistic(gain=1.0, bias=4.0),
#     #                                             # should be able to get same result with offset = -4.0
#     #                                             integrator_mode=True,
#     #                                             noise=pnl.NormalDist(mean=0, standard_deviation=unit_noise).function,
#     #                                             integration_rate=0.1,
#     #                                             name='COLORS HIDDEN')
#
#     colors_hidden_layer = pnl.TransferMechanism(size=2,
#                                                 function=pnl.Logistic(gain=1.0, x_0=4.0),
#                                                 # should be able to get same result with offset = -4.0
#                                                 integrator_mode=True,
#                                                 noise=0.0,
#                                                 integration_rate=0.1,
#                                                 name='COLORS HIDDEN')
#     #    words_hidden: ('RED','GREEN')
#     # words_hidden_layer = pnl.TransferMechanism(size=2,
#     #                                            function=pnl.Logistic(gain=1.0, bias=4.0),
#     #                                            integrator_mode=True,
#     #                                            noise=pnl.NormalDist(mean=0, standard_deviation=unit_noise).function,
#     #                                            integration_rate=0.1,
#     #                                            name='WORDS HIDDEN')
#     words_hidden_layer = pnl.TransferMechanism(size=2,
#                                                function=pnl.Logistic(gain=1.0, x_0=4.0),
#                                                integrator_mode=True,
#                                                noise=0.0,
#                                                integration_rate=0.1,
#                                                name='WORDS HIDDEN')
#
#     #    OUTPUT UNITS
#
#     #   Response layer, provide input to accumulator, responses: ('red', 'green')
#     #   time averaging = tau = 0.1
#     #   randomly distributed noise to the net input
#     # response_layer = pnl.TransferMechanism(size=2,
#     #                                        function=pnl.Logistic,
#     #                                        name='RESPONSE',
#     #                                        integrator_mode=True,
#     #                                        noise=pnl.NormalDist(mean=0, standard_deviation=unit_noise).function,
#     #                                        integration_rate=0.1)
#     response_layer = pnl.TransferMechanism(size=2,
#                                            function=pnl.Logistic,
#                                            name='RESPONSE',
#                                            integrator_mode=True,
#                                            noise=0.0,
#                                            integration_rate=0.1)
#     #   Respond red accumulator
#     #   alpha = rate of evidence accumlation = 0.1
#     #   sigma = noise = 0.1
#     #   noise will be: squareroot(time_step_size * noise) * a random sample from a normal distribution
#     accumulator_noise = 0.1
#     # respond_red_accumulator = pnl.IntegratorMechanism(function=pnl.SimpleIntegrator(noise=pnl.NormalDist(mean=0,
#     #                                                                                                      standard_deviation= accumulator_noise).function,
#     #                                                                                 rate=0.1),
#     #                                                   name='respond_red_accumulator')
#     respond_red_accumulator = pnl.IntegratorMechanism(function=pnl.SimpleIntegrator(noise=0.0,
#                                                                                     rate=0.1),
#                                                       name='respond_red_accumulator')
#     #   Respond green accumulator
#     # respond_green_accumulator = pnl.IntegratorMechanism(function=pnl.SimpleIntegrator(noise=pnl.NormalDist(mean=0,
#     #                                                                                                        standard_deviation=accumulator_noise).function,
#     #                                                                                   rate=0.1),
#     #                                                     name='respond_green_accumulator')
#     respond_green_accumulator = pnl.IntegratorMechanism(function=pnl.SimpleIntegrator(noise=0.0,
#                                                                                       rate=0.1),
#                                                         name='respond_green_accumulator')
#
#     #   LOGGING
#     colors_hidden_layer.set_log_conditions('value')
#     words_hidden_layer.set_log_conditions('value')
#     response_layer.set_log_conditions('value')
#     respond_red_accumulator.set_log_conditions('value')
#     respond_green_accumulator.set_log_conditions('value')
#
#     logged_mechanisms = [colors_hidden_layer, words_hidden_layer, response_layer, respond_red_accumulator, respond_green_accumulator]
#
#     #   SET UP CONNECTIONS
#
#     #   rows correspond to sender
#     #   columns correspond to: weighting of the contribution that a given sender makes to the receiver
#
#     #   INPUT TO HIDDEN
#     # row 0: input_'red' to hidden_'red', hidden_'green'
#     # row 1: input_'green' to hidden_'red', hidden_'green'
#     color_weights = pnl.MappingProjection(matrix=np.matrix([[2.2, -2.2],
#                                                             [-2.2, 2.2]]),
#                                           name='COLOR_WEIGHTS')
#     # row 0: input_'RED' to hidden_'RED', hidden_'GREEN'
#     # row 1: input_'GREEN' to hidden_'RED', hidden_'GREEN'
#     word_weights = pnl.MappingProjection(matrix=np.matrix([[2.6, -2.6],
#                                                            [-2.6, 2.6]]),
#                                          name='WORD_WEIGHTS')
#
#     #   HIDDEN TO RESPONSE
#     # row 0: hidden_'red' to response_'red', response_'green'
#     # row 1: hidden_'green' to response_'red', response_'green'
#     color_response_weights = pnl.MappingProjection(matrix=np.matrix([[1.3, -1.3],
#                                                                      [-1.3, 1.3]]),
#                                                    name='COLOR_RESPONSE_WEIGHTS')
#     # row 0: hidden_'RED' to response_'red', response_'green'
#     # row 1: hidden_'GREEN' to response_'red', response_'green'
#     word_response_weights = pnl.MappingProjection(matrix=np.matrix([[2.5, -2.5],
#                                                                     [-2.5, 2.5]]),
#                                                   name='WORD_RESPONSE_WEIGHTS')
#
#     #   TASK TO HIDDEN LAYER
#     #   row 0: task_CN to hidden_'red', hidden_'green'
#     #   row 1: task_WR to hidden_'red', hidden_'green'
#     task_CN_weights = pnl.MappingProjection(matrix=np.matrix([[4.0, 4.0],
#                                                               [0, 0]]),
#                                             name='TASK_CN_WEIGHTS')
#
#     #   row 0: task_CN to hidden_'RED', hidden_'GREEN'
#     #   row 1: task_WR to hidden_'RED', hidden_'GREEN'
#     task_WR_weights = pnl.MappingProjection(matrix=np.matrix([[0, 0],
#                                                               [4.0, 4.0]]),
#                                             name='TASK_WR_WEIGHTS')
#
#     #   RESPONSE UNITS TO ACCUMULATORS
#     #   row 0: response_'red' to respond_red_accumulator
#     #   row 1: response_'green' to respond_red_accumulator
#     respond_red_differencing_weights = pnl.MappingProjection(matrix=np.matrix([[1.0], [-1.0]]),
#                                                              name='RESPOND_RED_WEIGHTS')
#
#     #   row 0: response_'red' to respond_green_accumulator
#     #   row 1: response_'green' to respond_green_accumulator
#     respond_green_differencing_weights = pnl.MappingProjection(matrix=np.matrix([[-1.0], [1.0]]),
#                                                                name='RESPOND_GREEN_WEIGHTS')
#
#     #   CREATE PATHWAYS
#     #   Words pathway
#     words_process = pnl.Process(pathway=[words_input_layer,
#                                          word_weights,
#                                          words_hidden_layer,
#                                          word_response_weights,
#                                          response_layer], name='WORDS_PROCESS')
#
#     #   Colors pathway
#     colors_process = pnl.Process(pathway=[colors_input_layer,
#                                           color_weights,
#                                           colors_hidden_layer,
#                                           color_response_weights,
#                                           response_layer], name='COLORS_PROCESS')
#
#     #   Task representation pathway
#     task_CN_process = pnl.Process(pathway=[task_layer,
#                                            task_CN_weights,
#                                            colors_hidden_layer],
#                                   name='TASK_CN_PROCESS')
#     task_WR_process = pnl.Process(pathway=[task_layer,
#                                            task_WR_weights,
#                                            words_hidden_layer],
#                                   name='TASK_WR_PROCESS')
#
#     #   Evidence accumulation pathway
#     respond_red_process = pnl.Process(pathway=[response_layer,
#                                                respond_red_differencing_weights,
#                                                respond_red_accumulator],
#                                       name='RESPOND_RED_PROCESS')
#     respond_green_process = pnl.Process(pathway=[response_layer,
#                                                  respond_green_differencing_weights,
#                                                  respond_green_accumulator],
#                                         name='RESPOND_GREEN_PROCESS')
#
#     #   CREATE SYSTEM
#     my_Stroop = pnl.System(processes=[colors_process,
#                                       words_process,
#                                       task_CN_process,
#                                       task_WR_process,
#                                       respond_red_process,
#                                       respond_green_process],
#                            name='FEEDFORWARD_STROOP_SYSTEM')
#
#     # my_Stroop.show()
#     # my_Stroop.show_graph(show_dimensions=pnl.ALL)
#
#     # Function to create test trials
#     # a RED word input is [1,0] to words_input_layer and GREEN word is [0,1]
#     # a red color input is [1,0] to colors_input_layer and green color is [0,1]
#     # a color-naming trial is [1,0] to task_layer and a word-reading trial is [0,1]
#
#     def trial_dict(red_color, green_color, red_word, green_word, CN, WR):
#
#         trialdict = {
#             colors_input_layer: [[0, 0], [red_color, green_color]],
#             words_input_layer: [[0, 0], [red_word, green_word]],
#             task_layer: [[CN, WR], [CN, WR]]
#         }
#         return trialdict
#
#     #   CREATE THRESHOLD FUNCTION
#     # first value of DDM's value is DECISION_VARIABLE
#     def pass_threshold(mech1, mech2, thresh):
#         results1 = mech1.output_ports[0].value
#         results2 = mech2.output_ports[0].value
#         for val in results1:
#             if val >= thresh:
#                 return True
#         for val in results2:
#             if val >= thresh:
#                 return True
#         return False
#
#     accumulator_threshold = 1.0
#
#     mechanisms_to_update = [colors_hidden_layer, words_hidden_layer, response_layer]
#
#     def switch_integrator_mode(mechanisms, mode):
#         for mechanism in mechanisms:
#             mechanism.integrator_mode = mode
#
#     def switch_noise(mechanisms, noise):
#         for mechanism in mechanisms:
#             mechanism.noise = noise
#
#     def switch_to_initialization_trial(mechanisms):
#         # Turn off accumulation
#         switch_integrator_mode(mechanisms, False)
#         # Turn off noise
#         switch_noise(mechanisms, 0)
#         # Execute once per trial
#         my_Stroop.termination_processing = {pnl.TimeScale.TRIAL: pnl.AllHaveRun()}
#
#     def switch_to_processing_trial(mechanisms):
#         # Turn on accumulation
#         switch_integrator_mode(mechanisms, True)
#         # Turn on noise
#         # switch_noise(mechanisms, pnl.NormalDist(mean=0, standard_deviation=unit_noise).function)
#         # Execute until one of the accumulators crosses the threshold
#         my_Stroop.termination_processing = {pnl.TimeScale.TRIAL: pnl.While(pass_threshold,
#                                                                            respond_red_accumulator,
#                                                                            respond_green_accumulator,
#                                                                            accumulator_threshold)}
#
#     def switch_trial_type():
#         # Next trial will be a processing trial
#         if isinstance(my_Stroop.termination_processing[pnl.TimeScale.TRIAL], pnl.AllHaveRun):
#             switch_to_processing_trial(mechanisms_to_update)
#         # Next trial will be an initialization trial
#         else:
#             switch_to_initialization_trial(mechanisms_to_update)
#
#     def _extract_rt_cycles(mechanism):
#         # Grab the log dictionary from the output layer
#         log_dict = mechanism.log.nparray_dictionary()
#
#         # Extract out the relevant keys from the log to a single numpy array
#         relevant_key_arrays = [np.array([x[0] for x in log_dict[key]]) for key in ('Run', 'Trial', 'Pass')]
#         table = np.stack(relevant_key_arrays, axis=1)
#
#         # Filter out only the last run
#         last_run = np.max(table[:, 0])
#         table = table[table[:, 0] == last_run]
#
#         # Filter out only the last pass of each trial
#         trial_ends = (table[1:, 1] - table[:-1, 1]) != 0
#         trial_ends = np.append(trial_ends, True)
#         last_passes = table[trial_ends, :]
#
#         # Filter out only odd trials
#         last_passes = last_passes[last_passes[:, 1] % 2 == 1, :]
#         return last_passes[:, 2]
#
#     def last_run_to_dataframe(mechanism_list):
#         dataframes = []
#         first = True
#         for log_layer in mechanism_list:
#             layer_size = log_layer.size[0]
#             log_dict = log_layer.log.nparray_dictionary()
#
#             # Extract out all keys, treating value specially since it's already an np array
#             arrays = [np.array([x[0] for x in log_dict[key]]) for key in ('Run', 'Trial', 'Pass', 'Time_step')]
#             arrays.extend([np.squeeze(log_dict['value'][:, :, i]) for i in range(layer_size)])
#             table = np.stack(arrays, axis=1)
#
#             # Filter out only the last run
#             last_run = np.max(table[:, 0])
#             table = table[table[:, 0] == last_run]
#
#             # Create as dataframe and add to the list of dataframes
#             if first:
#                 df = pandas.DataFrame(table, columns=['Run', 'Trial', 'Pass', 'Time_step'] +
#                                                      [f'{log_layer.name}_{i}' for i in range(layer_size)])
#                 first = False
#
#             else:
#                 df = pandas.DataFrame(table[:, -1 * layer_size:], columns=[f'{log_layer.name}_{i}'
#                                                                            for i in range(layer_size)])
#
#             dataframes.append(df)
#
#         return pandas.concat(dataframes, axis=1, join='inner')
#
#
#     # Start with an initialization trial
#     switch_to_initialization_trial(mechanisms_to_update)
#
#     my_Stroop.run(inputs=trial_dict(red_color, green_color, red_word, green_word, CN, WR),
#                   # termination_processing=change_termination_processing,
#                   num_trials=n_trials,
#                   call_after_trial=switch_trial_type)
#
#     # respond_red_accumulator.log.print_entries()
#     # respond_green_accumulator.log.print_entries()
#     # response_layer.log.print_entries()
#     my_Stroop_rt_cycles = _extract_rt_cycles(respond_green_accumulator)
#     my_Stroop_DataFrame = last_run_to_dataframe(logged_mechanisms)
#     # print(my_Stroop_DataFrame)
#     respond_red_accumulator.log.print_entries
#
#     return my_Stroop_rt_cycles
#
# incong_results = test_lauras_cohen_1990_model(0, 1, 1, 0, 1, 0, 10)
#
# print(incong_results)
#
# plt.hist(incong_results)
# plt.title('Reaction times for incongruent CN trial where green is correct response')
# plt.xlabel("number of cycles")
# plt.ylabel("number of trials out of 100")
# plt.show()
#
# #test WR control trial
# results_WR_control_trial = test_lauras_cohen_1990_model(0, 0, 1, 0, 0, 1, 50)
#
# #test WR congruent trial (should have the least cycles)
# results_WR_congruent_trial = test_lauras_cohen_1990_model(1, 0, 1, 0, 0, 1, 50)
#
# #test WR incongruent trial, should see that color doesn't affect word (same number of cycles as WR control)
# results_WR_incongruent_trial = test_lauras_cohen_1990_model(0, 1, 1, 0, 0, 1, 50)
#
# #test CN control trial
# results_CN_control_trial = test_lauras_cohen_1990_model(1, 0, 0, 0, 1, 0, 50)
#
# #test CN congruent trial (should have more cycles than WR congruent)
# results_CN_congruent_trial = test_lauras_cohen_1990_model(1, 0, 1, 0, 1, 0, 50)
#
# #test CN incongruent trial, should see that word interferes with color (should have most cycles + more than CN control)
# results_CN_incongruent_trial = test_lauras_cohen_1990_model(1, 0, 0, 1, 1, 0, 50)
#
#
# print(results_CN_incongruent_trial)
# cycles_mean = [np.mean(results_WR_control_trial[0]),
#      np.mean(results_WR_incongruent_trial[0]),
#      np.mean(results_WR_congruent_trial[0]),
#      np.mean(results_CN_control_trial[0]),
#      np.mean(results_CN_incongruent_trial[0]),
#      np.mean(results_CN_congruent_trial[0])]
# cycles_std = [np.std(results_WR_control_trial),
#          np.std(results_WR_incongruent_trial),
#          np.std(results_WR_congruent_trial),
#          np.std(results_CN_control_trial),
#          np.std(results_CN_incongruent_trial),
#          np.std(results_CN_congruent_trial)]
# cycles_x = np.array([0, 1, 2, 0, 1, 2])
# labs = ['control',
#         'conflict',
#         'congruent']
# legend = ['WR trial',
#           'CN trial']
# colors = ['b', 'c']
#
# print(np.mean(results_WR_congruent_trial[0]))
# print('of 0',np.mean(results_WR_congruent_trial))
#
# plt.plot(cycles_x[0:3], cycles_mean[0:3], color=colors[0])
# plt.errorbar(cycles_x[0:3], cycles_mean[0:3], xerr=0, yerr=cycles_std[0:3], ecolor=colors[0], fmt='none')
# plt.scatter(cycles_x[0], cycles_mean[0], marker='x', color=colors[0])
# plt.scatter(cycles_x[1], cycles_mean[1], marker='x', color=colors[0])
# plt.scatter(cycles_x[2], cycles_mean[2], marker='x', color=colors[0])
# plt.plot(cycles_x[3:6], cycles_mean[3:6], color=colors[1])
# plt.errorbar(cycles_x[3:6], cycles_mean[3:6], xerr=0, yerr=cycles_std[3:6], ecolor=colors[1], fmt='none')
# plt.scatter(cycles_x[3], cycles_mean[3], marker='o', color=colors[1])
# plt.scatter(cycles_x[4], cycles_mean[4], marker='o', color=colors[1])
# plt.scatter(cycles_x[5], cycles_mean[5], marker='o', color=colors[1])
#
# plt.xticks(cycles_x, labs, rotation=15)
# plt.tick_params(axis='x', labelsize=9)
# plt.title('Mean Number of Cycles by trial type')
# plt.legend(legend)
# plt.show()


import timeit
import numpy as np
from psyneulink import *

from gym_forager.envs.forager_env import ForagerEnv

# Runtime Switches:
RENDER = False
PNL_COMPILE = False
PERCEPTUAL_DISTORT = False
RUN = False
SHOW_GRAPH = True


# *********************************************************************************************************************
# *********************************************** CONSTANTS ***********************************************************
# *********************************************************************************************************************

# These should probably be replaced by reference to ForagerEnv constants:
obs_len = 3
obs_coords = 2
action_len = 2
player_idx = 0
player_obs_start_idx = player_idx * obs_len
player_value_idx = player_idx * obs_len + obs_coords
player_coord_slice = slice(player_obs_start_idx,player_value_idx)
predator_idx = 1
predator_obs_start_idx = predator_idx * obs_len
predator_value_idx = predator_idx * obs_len + obs_coords
predator_coord_slice = slice(predator_obs_start_idx,predator_value_idx)
prey_idx = 2
prey_obs_start_idx = prey_idx * obs_len
prey_value_idx = prey_idx * obs_len + obs_coords
prey_coord_slice = slice(prey_obs_start_idx,prey_value_idx)

player_len = prey_len = predator_len = obs_coords


# *********************************************************************************************************************
# **************************************  MECHANISMS AND COMPOSITION  *************************************************
# *********************************************************************************************************************

# Input Mechanisms
player_input = ProcessingMechanism(size=prey_len, name="PLAYER INPUT")
prey_input = ProcessingMechanism(size=prey_len, name="PREY INPUT")
predator_input = TransferMechanism(size=predator_len, name="PREDATOR INPUT")

# Perceptual Mechanisms
if PERCEPTUAL_DISTORT:
    player_obs = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PLAYER OBS")
    prey_obs = ProcessingMechanism(size=prey_len, function=GaussianDistort, name="PREY OBS")
    predator_obs = TransferMechanism(size=predator_len, function=GaussianDistort, name="PREDATOR OBS")
else:
    player_obs = ProcessingMechanism(size=prey_len, name="PLAYER OBS")
    prey_obs = ProcessingMechanism(size=prey_len, name="PREY OBS")
    predator_obs = TransferMechanism(size=predator_len, name="PREDATOR OBS")


# Value and Reward Mechanisms (not yet used;  for future use)
values = TransferMechanism(size=3, name="AGENT VALUES")
reward = TransferMechanism(name="REWARD")

# Action Mechanism
#    Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
#    note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
greedy_action_mech = ComparatorMechanism(name='ACTION',sample=player_obs,target=prey_obs)

# Create Composition
agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')
agent_comp.add_node(player_obs)
agent_comp.add_node(predator_obs)
agent_comp.add_node(prey_obs)
# agent_comp.add_linear_processing_pathway([player_input, player_obs])
# agent_comp.add_linear_processing_pathway([predator_input, predator_obs])
# agent_comp.add_linear_processing_pathway([prey_input, prey_obs])
agent_comp.add_node(greedy_action_mech)

# ControlMechanism

#   function for ObjectiveMechanism
def diff_fct(variable):
    # Get difference in distance of player to predator vs. prey
    if variable is None:
        return 0
    player_coord = variable[0]
    player_percept = variable[1]
    predator_coord = variable[2]
    predator_percept = variable[3]
    prey_coord = variable[4]
    prey_percept = variable[5]
    player_diff = np.sum(np.abs(player_percept - player_coord))
    predator_diff = np.sum(np.abs(predator_percept - predator_coord))
    prey_diff = np.sum(np.abs(prey_percept - prey_coord))
    # return - (np.sum(player_diff) + np.sum(predator_diff))
    return -(np.sum(player_diff))

def test_fct(variable):
    if variable is None:
        return 0
    return variable[1] - variable[0]

if PERCEPTUAL_DISTORT:
    CTL_PARAM = VARIANCE
else:
    CTL_PARAM = SLOPE
# agent_comp._analyze_graph()
ocm = OptimizationControlMechanism(features=[player_obs.input_port, predator_obs.input_port, prey_obs.input_port],
                                   agent_rep=agent_comp,
                                   function=GridSearch(direction=MAXIMIZE,
                                                       save_values=True),
                                   objective_mechanism=ObjectiveMechanism(
                                           # function=diff_fct,
                                           function=test_fct,
                                           monitor=[player_obs,
                                                    player_obs.input_port,
                                                    predator_obs,
                                                    predator_obs.input_port,
                                                    prey_obs,
                                                    prey_obs.input_port
                                                    ]
                                           # monitored_output_ports=[player_input, player_obs,
                                           #                          predator_input, predator_obs,
                                           #                          prey_input, prey_obs
                                           #                          ]
                                           # monitored_output_ports=[agent_comp.input_CIM_ports[
                                           #                              player_obs.input_port][1],
                                           #                          player_obs,
                                           #                          agent_comp.input_CIM_ports[
                                           #                              predator_obs.input_port][1],
                                           #                          predator_obs,
                                           #                          agent_comp.input_CIM_ports[
                                           #                              prey_obs.input_port][1],
                                           #                          prey_obs
                                           #                          ]
                                   ),
                                   control_signals=[ControlSignal(modulates=(CTL_PARAM, player_obs),
                                                                  # allocation_samples=[0, 1, 10, 100]),
                                                                  # allocation_samples=[0, 10, 100]),
                                                                  # allocation_samples=[10, 1]),
                                                                  allocation_samples=[0, 100],
                                                                  intensity_cost_function=Exponential(rate=-.1,
                                                                                                      bias=5),
                                                                  ),
                                                    ControlSignal(modulates=(CTL_PARAM, predator_obs),
                                                                  # allocation_samples=[0, 1, 10, 100]),
                                                                  # allocation_samples=[0, 10, 100]),
                                                                  # allocation_samples=[10, 1]),
                                                                  allocation_samples=[0, 100],
                                                                  intensity_cost_function=Exponential(rate=-.1,
                                                                                                      bias=5),
                                                                  ),
                                                    ControlSignal(modulates=(CTL_PARAM, prey_obs),
                                                                  # allocation_samples=[0, 1, 10, 100]),
                                                                  # allocation_samples=[0, 10, 100]),
                                                                  # allocation_samples=[10, 1]),
                                                                  allocation_samples=[0, 100],
                                                                  intensity_cost_function=Exponential(rate=-.1,
                                                                                                      bias=5),
                                                                  ),
                                                    ],
                                   )
agent_comp.add_model_based_optimizer(ocm)
agent_comp.enable_model_based_optimizer = True

full_comp = Composition(name='FULL_COMPOSITION')
full_comp.add_node(agent_comp)
full_comp.add_node(player_input)
full_comp.add_node(predator_input)
full_comp.add_node(prey_input)

# full_comp.add_projection(sender=player_input, receiver=player_obs)
# full_comp.add_projection(sender=predator_input, receiver=predator_obs)
# full_comp.add_projection(sender=prey_input, receiver=prey_obs)

full_comp.add_linear_processing_pathway([player_input,player_obs])
full_comp.add_linear_processing_pathway([predator_input,predator_obs])
full_comp.add_linear_processing_pathway([prey_input,prey_obs])

if SHOW_GRAPH:
    # agent_comp.show_graph(show_mechanism_structure='ALL')
    # agent_comp.show_graph(show_controller=True)
    full_comp.show_graph(show_controller=True)


# *********************************************************************************************************************
# ******************************************   RUN SIMULATION  ********************************************************
# *********************************************************************************************************************

num_trials = 1

def main():
    env = ForagerEnv()
    reward = 0
    done = False
    if RENDER:
        env.render()  # If visualization is desired
    else:
        print("Running simulation...")
    steps = 0
    start_time = timeit.default_timer()
    for _ in range(num_trials):
        observation = env.reset()
        while True:
            if PNL_COMPILE:
                BIN_EXECUTE = 'LLVM'
            else:
                BIN_EXECUTE = 'Python'

            run_results = full_comp.run(inputs={player_input:[observation[player_coord_slice]],
                                                 predator_input:[observation[predator_coord_slice]],
                                                 prey_input:[observation[prey_coord_slice]],
                                                 },
                                         bin_execute=BIN_EXECUTE
                                         )
            action = np.where(run_results[0] == 0, 0, run_results[0] / np.abs(run_results[0]))
            # action = np.squeeze(np.where(greedy_action_mech.value==0,0,
            #                              greedy_action_mech.value[0]/np.abs(greedy_action_mech.value[0])))
            observation, reward, done, _ = env.step(action)
            print('\nStep: ', steps)
            print('Outcome: {}'.format(ocm.objective_mechanism.value))
            print('OCM ControlSignals:')
            print('\tPlayer OBS: {}\n\tPredator OBS: {}\n\tPrey OBS: {}'.
                  format(ocm.control_signals[0].value,
                         ocm.control_signals[1].value,
                         ocm.control_signals[2].value))
            print('OCM ControlSignal Costs:')
            print('\tPlayer OBS: {}\n\tPredator OBS: {}\n\tPrey OBS: {}'.
                  format(ocm.control_signals[0].cost,
                         ocm.control_signals[1].cost,
                         ocm.control_signals[2].cost))
            print('SIMULATION (PREP FOR NEXT TRIAL):')
            for sample, value in zip(ocm.saved_samples, ocm.saved_values):
                print('\t\tSample: {} Value: {}'.format(sample, value))
            steps += 1
            if done:
                break
    stop_time = timeit.default_timer()
    print(f'{steps / (stop_time - start_time):.1f} steps/second, {steps} total steps in '
          f'{stop_time - start_time:.2f} seconds')
    if RENDER:
        env.render()  # If visualization is desired

if RUN:
    if __name__ == "__main__":
        main()



import numpy as np
import pytest

from psyneulink.core.components.functions.statefulfunctions.memoryfunctions import ContentAddressableMemory
from psyneulink.library.components.mechanisms.processing.integrator.episodicmemorymechanism import EpisodicMemoryMechanism
import psyneulink.core.llvm as pnlvm

np.random.seed(0)
CONTENT_SIZE=10
ASSOC_SIZE=10
test_var = np.random.rand(2, CONTENT_SIZE)
test_initializer = {tuple(test_var[0]): test_var[1]}


test_data = [
    (test_var, ContentAddressableMemory, {'default_variable':test_var}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047, 0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777], [
       0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694, 0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
    (test_var, ContentAddressableMemory, {'default_variable':test_var, 'retrieval_prob':0.5},
       [[ 0. for i in range(CONTENT_SIZE) ],[ 0. for i in range(ASSOC_SIZE) ]]),
    (test_var, ContentAddressableMemory, {'default_variable':test_var, 'storage_prob':0.1},
       [[ 0. for i in range(CONTENT_SIZE) ],[ 0. for i in range(ASSOC_SIZE) ]]),
    (test_var, ContentAddressableMemory, {'default_variable':test_var, 'retrieval_prob':0.9, 'storage_prob':0.9}, [[
       0.5488135039273248, 0.7151893663724195, 0.6027633760716439, 0.5448831829968969, 0.4236547993389047, 0.6458941130666561, 0.4375872112626925, 0.8917730007820798, 0.9636627605010293, 0.3834415188257777], [
       0.7917250380826646, 0.5288949197529045, 0.5680445610939323, 0.925596638292661, 0.07103605819788694, 0.08712929970154071, 0.02021839744032572, 0.832619845547938, 0.7781567509498505, 0.8700121482468192 ]]),
]

# use list, naming function produces ugly names
names = [
    "ContentAddressableMemory",
    "ContentAddressableMemory Random Retrieval",
    "ContentAddressableMemory Random Storage",
    "ContentAddressableMemory Random Retrieval-Storage",
]

@pytest.mark.function
@pytest.mark.memory_function
@pytest.mark.benchmark
@pytest.mark.parametrize('variable, func, params, expected', test_data, ids=names)
@pytest.mark.parametrize('mode', ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_basic(variable, func, params, expected, benchmark, mode):
    f = func(seed=0, **params)
    m = EpisodicMemoryMechanism(content_size=len(variable[0]), assoc_size=len(variable[1]), function=f)
    if mode == 'Python':
        def EX(variable):
            m.execute(variable)
            return m.output_values
    elif mode == 'LLVM':
        EX = pnlvm.execution.MechExecution(m).execute
    elif mode == 'PTX':
        EX = pnlvm.execution.MechExecution(m).cuda_execute

    EX(variable)
    res = EX(variable)
    assert np.allclose(res[0], expected[0])
    assert np.allclose(res[1], expected[1])
    if benchmark.enabled:
        benchmark(EX, variable)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **********************************************  Mechanism ***********************************************************

"""
**[DOCUMENTATION STILL UNDER CONSTRUCTION]**

"""
import numpy as np
import typecheck as tc

from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.mechanisms.mechanism import Mechanism_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.defaults import SystemDefaultInputValue
from psyneulink.core.globals.keywords import DEFAULT_PROCESSING_MECHANISM, FUNCTION, FUNCTION_PARAMS, INTERCEPT, SLOPE
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

# **************************************** DefaultProcessingMechanism ******************************************************

__all__ = []


# class DefaultProcessingMechanism_Base(ProcessingMechanism_Base):
class DefaultProcessingMechanism_Base(Mechanism_Base):
    """Subclass of `ProcessingMechanism <ProcessingMechanism>` used to implement SystemDefaultInputMechanism,
    DefaultControlMechanism, and SystemDefaultOutputMechanism.

    Description:
        Implements "dummy" Mechanism used to implement default input, control signals, and outputs to other mechanisms

    Class attributes:
        + componentType (str): System Default Mechanism
    """

    componentName = DEFAULT_PROCESSING_MECHANISM
    componentType = DEFAULT_PROCESSING_MECHANISM
    onlyFunctionOnInit = True

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # Any preferences specified below will override those specified in SUBTYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to SUBTYPE automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'DefaultProcessingMechanismClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    class Parameters(Mechanism_Base.Parameters):
        variable = np.array([SystemDefaultInputValue])

    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 size=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 function=None,
                 **kwargs
                 ):
        """Add Linear as default function, assign default name, and call super.__init__

        :param default_variable: (value)
        :param size: (int or list/array of ints)
        :param params: (dict)
        :param name: (str)
        :param prefs: (PreferenceSet)
        """

        super(DefaultProcessingMechanism_Base, self).__init__(default_variable=default_variable,
                                                              size=size,
                                                              function=function,
                                                              params=params,
                                                              name=name,
                                                              prefs=prefs,

                                                              **kwargs)



import numpy as np
import psyneulink.core.llvm as pnlvm
import psyneulink.core.components.functions.statefulfunctions.integratorfunctions
import pytest

np.random.seed(0)
params = {'a_v' : np.random.rand(),
          'b_v' : np.random.rand(),
          'c_v' : np.random.rand(),
          'd_v' : np.random.rand(),
          'e_v' : np.random.rand(),
          'f_v' : np.random.rand(),
          'a_w' : np.random.rand(),
          'b_w' : np.random.rand(),
          'c_w' : np.random.rand(),
          'time_constant_v' : np.random.rand(),
          'time_constant_w' : np.random.rand(),
          'threshold' : np.random.rand(),
          'uncorrelated_activity' : np.random.rand(),
          'mode' : np.random.rand(),
          }

SIZE=8
test_var = np.random.rand(SIZE)
test_scalar = np.random.rand()

test_data = [
    (psyneulink.core.components.functions.statefulfunctions.integratorfunctions.FitzHughNagumoIntegrator, test_var, "RK4", params, ([0.23619944, 0.24032298, 0.22321782, 0.43865125, 0.42363054,
                                                                                                                          0.44901757, 0.47938108, 0.42941189], [0.21378097, 0.21388886, 0.21344061, 0.21894107, 0.21856817,
       0.21919746, 0.21994384, 0.2187119 ], 0.15000000000000002)),
    (psyneulink.core.components.functions.statefulfunctions.integratorfunctions.FitzHughNagumoIntegrator, test_scalar, "RK4", params,
     ([0.33803257], [0.21641212], 0.15000000000000002)),
    (psyneulink.core.components.functions.statefulfunctions.integratorfunctions.FitzHughNagumoIntegrator, test_var, "EULER", params, ([0.23686576, 0.24093183, 0.22404678, 0.43291206, 0.41863405,
                                                                                                                            0.44273909,
                                                                                                                            0.47139546, 0.42413492], [0.20757016, 0.2076755 , 0.20723764, 0.21257185, 0.21221299,
       0.21281834, 0.21353476, 0.21235135], 0.15000000000000002)),
    (psyneulink.core.components.functions.statefulfunctions.integratorfunctions.FitzHughNagumoIntegrator, test_scalar, "EULER", params, ([0.33642314], [0.21013003], 0.15000000000000002)),
]

# use list, naming function produces ugly names
names = [
    "FitzHughNagumoIntegrator RK4 VECTOR",
    "FitzHughNagumoIntegrator RK4 SCALAR",
    "FitzHughNagumoIntegrator EULER VECTOR",
    "FitzHughNagumoIntegrator EULER SCALAR",
]

@pytest.mark.function
@pytest.mark.integrator_function
@pytest.mark.fitzHughNagumo_integrator_function
@pytest.mark.benchmark(group="FitzHughNagumoIntegrator")
@pytest.mark.parametrize("func, variable, integration_method, params, expected", test_data, ids=names)
@pytest.mark.parametrize('mode', ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_basic(func, variable, integration_method, params, expected, benchmark, mode):
    f = func(default_variable=variable, integration_method=integration_method, params=params)
    if mode == 'Python':
        EX = f.function
    elif mode == 'LLVM':
        e = pnlvm.execution.FuncExecution(f)
        EX = e.execute
    elif mode == 'PTX':
        e = pnlvm.execution.FuncExecution(f)
        EX = e.cuda_execute

    res = EX(variable)
    res = EX(variable)
    res = EX(variable)

    assert np.allclose(res[0], expected[0])
    assert np.allclose(res[1], expected[1])
    assert np.allclose(res[2], expected[2])

    if benchmark.enabled:
        benchmark(f, variable)



import pytest
import doctest


import psyneulink as pnl


@pytest.mark.parametrize("mod", [# Ports
                                 pnl.core.components.ports.parameterport,
                                 pnl.core.components.ports.outputport,
                                 pnl.core.components.ports.modulatorysignals.controlsignal,
                                 pnl.core.components.ports.modulatorysignals.gatingsignal,
                                 # Mechanisms
                                 # FIX 5/8/20 ELIMINATE SYSTEM [JDC] -- REFERENCES TO LABELS REQUIRE REFACTORING
                                 # pnl.core.components.mechanisms.mechanism,
                                 pnl.core.components.mechanisms.processing.transfermechanism,
                                 pnl.core.components.mechanisms.processing.integratormechanism,
                                 pnl.core.components.mechanisms.processing.objectivemechanism,
                                 pnl.core.components.mechanisms.modulatory.control.controlmechanism,
                                 # Function
                                 pnl.core.components.functions.function,
                                ])
def test_core_docs(mod, capsys):
    fail, total = doctest.testmod(mod)
    if fail > 0:
        captured = capsys.readouterr()
        pytest.fail("{} out of {} examples failed:\n{}\n{}".format(
            fail, total, captured.err, captured.out), pytrace=False)

@pytest.mark.parametrize("mod", [# Mechanisms
                                 pnl.library.components.mechanisms.processing.integrator.ddm,
                                 pnl.library.components.mechanisms.processing.objective.comparatormechanism,
                                 # Scheduling
                                 pnl.core.scheduling.scheduler,
                                 # Logs
                                 pnl.core.globals.log,
                                ])
def test_other_docs(mod, capsys):
    fail, total = doctest.testmod(mod, optionflags=doctest.REPORT_NDIFF)
    if fail > 0:
        captured = capsys.readouterr()
        pytest.fail("{} out of {} examples failed:\n{}\n{}".format(
            fail, total, captured.err, captured.out), pytrace=False)



import numpy as np
import pytest

import psyneulink as pnl

# This script implements Figure 1 of Botvinick, M. M., Braver, T. S., Barch, D. M., Carter, C. S., & Cohen, J. D. (2001).
# Conflict monitoring and cognitive control. Psychological Review, 108, 624–652.
# http://dx.doi.org/10.1037/0033-295X.108.3.624

# Figure 1 plots the ENERGY computed by a conflict mechanism. It is highest for incongruent trials,
# and similar for congruent and neutral trials.
# Noise is turned of and for each condition we ran one trial only. A response threshold was not defined. Responses were
# made at the marked * signs in the figure.
# Note that this script implements a slightly different Figure than in the original Figure in the paper.
# However, this implementation is identical with a plot we created with an old MATLAB code which was used for the
# conflict monitoring simulations.
import psyneulink.core.components.functions.objectivefunctions
import psyneulink.core.components.functions.transferfunctions


@pytest.mark.model
@pytest.mark.benchmark
@pytest.mark.parametrize("reps", [1,
                                  pytest.param(10, marks=pytest.mark.stress),
                                  pytest.param(100, marks=pytest.mark.stress)])
@pytest.mark.parametrize("mode", ['Python',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.param('LLVMExec', marks=pytest.mark.llvm),
                                  pytest.param('LLVMRun', marks=pytest.mark.llvm),
                                  pytest.param('PTXExec', marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                  pytest.param('PTXRun', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_botvinick_model(benchmark, mode, reps):
    benchmark.group = "Botvinick (scale " + str(reps / 100) + ")"

    # SET UP MECHANISMS ----------------------------------------------------------------------------------------------------
    # Linear input layer
    # colors: ('red', 'green'), words: ('RED','GREEN')
    colors_input_layer = pnl.TransferMechanism(size=3,
                                               function=psyneulink.core.components.Linear,
                                               name='COLORS_INPUT')

    words_input_layer = pnl.TransferMechanism(size=3,
                                              function=psyneulink.core.components.Linear,
                                              name='WORDS_INPUT')

    task_input_layer = pnl.TransferMechanism(size=2,
                                             function=psyneulink.core.components.Linear,
                                             name='TASK_INPUT')

    #   Task layer, tasks: ('name the color', 'read the word')
    task_layer = pnl.RecurrentTransferMechanism(size=2,
                                                function=psyneulink.core.components.Logistic,
                                                hetero=-2,
                                                integrator_mode=True,
                                                integration_rate=0.01,
                                                name='TASK_LAYER')

    # Hidden layer
    # colors: ('red','green', 'neutral') words: ('RED','GREEN', 'NEUTRAL')
    colors_hidden_layer = pnl.RecurrentTransferMechanism(size=3,
                                                         function=psyneulink.core.components.Logistic(x_0=4.0),  # bias 4.0 is -4.0 in the paper see Docs for description
                                                         integrator_mode=True,
                                                         hetero=-2,
                                                         integration_rate=0.01,  # cohen-huston text says 0.01
                                                         name='COLORS_HIDDEN')

    words_hidden_layer = pnl.RecurrentTransferMechanism(size=3,
                                                        function=psyneulink.core.components.Logistic(x_0=4.0),
                                                        integrator_mode=True,
                                                        hetero=-2,
                                                        integration_rate=0.01,
                                                        name='WORDS_HIDDEN')

    #   Response layer, responses: ('red', 'green')
    response_layer = pnl.RecurrentTransferMechanism(size=2,
                                                    function=psyneulink.core.components.Logistic,
                                                    hetero=-2.0,
                                                    integrator_mode=True,
                                                    integration_rate=0.01,
                                                    output_ports = [pnl.RESULT,
                                                                     {pnl.NAME: 'DECISION_ENERGY',
                                                                      pnl.VARIABLE: (pnl.OWNER_VALUE,0),
                                                                      pnl.FUNCTION: psyneulink.core.components.Stability(
                                                                          default_variable = np.array([0.0, 0.0]),
                                                                          metric = pnl.ENERGY,
                                                                          matrix = np.array([[0.0, -4.0],
                                                                                            [-4.0, 0.0]]))}],
                                                    name='RESPONSE', )

    # Mapping projections---------------------------------------------------------------------------------------------------

    color_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0, 0.0],
                                                                 [0.0, 1.0, 0.0],
                                                                 [0.0, 0.0, 1.0]]))

    word_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0, 0.0],
                                                                [0.0, 1.0, 0.0],
                                                                [0.0, 0.0, 1.0]]))

    task_input_weights = pnl.MappingProjection(matrix=np.array([[1.0, 0.0],
                                                                [0.0, 1.0]]))

    color_task_weights  = pnl.MappingProjection(matrix=np.array([[4.0, 0.0],
                                                                 [4.0, 0.0],
                                                                 [4.0, 0.0]]))

    task_color_weights  = pnl.MappingProjection(matrix=np.array([[4.0, 4.0, 4.0],
                                                                 [0.0, 0.0, 0.0]]))

    response_color_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0, 0.0],
                                                                    [0.0, 1.5, 0.0]]))

    response_word_weights  = pnl.MappingProjection(matrix=np.array([[2.5, 0.0, 0.0],
                                                                    [0.0, 2.5, 0.0]]))

    color_response_weights = pnl.MappingProjection(matrix=np.array([[1.5, 0.0],
                                                                    [0.0, 1.5],
                                                                    [0.0, 0.0]]))

    word_response_weights  = pnl.MappingProjection(matrix=np.array([[2.5, 0.0],
                                                                    [0.0, 2.5],
                                                                    [0.0, 0.0]]))

    word_task_weights = pnl.MappingProjection(matrix=np.array([[0.0, 4.0],
                                                               [0.0, 4.0],
                                                               [0.0, 4.0]]))

    task_word_weights = pnl.MappingProjection(matrix=np.array([[0.0, 0.0, 0.0],
                                                               [4.0, 4.0, 4.0]]))

    # CREATE Composition
    comp = pnl.Composition()

    # Add mechanisms
    comp.add_node(colors_input_layer)
    comp.add_node(colors_hidden_layer)

    comp.add_node(words_input_layer)
    comp.add_node(words_hidden_layer)

    comp.add_node(task_input_layer)
    comp.add_node(task_layer)
    comp.add_node(response_layer)

    # Add projections
    comp.add_projection(task_input_weights, task_input_layer, task_layer)

    # Color process
    comp.add_projection(color_input_weights, colors_input_layer, colors_hidden_layer)
    comp.add_projection(color_response_weights, colors_hidden_layer, response_layer)
    comp.add_projection(response_color_weights, response_layer, colors_hidden_layer)

    # Word process
    comp.add_projection(word_input_weights, words_input_layer, words_hidden_layer)
    comp.add_projection(word_response_weights, words_hidden_layer, response_layer)
    comp.add_projection(response_word_weights, response_layer, words_hidden_layer)

    # Color task process
    comp.add_projection(task_color_weights, task_layer, colors_hidden_layer)
    comp.add_projection(color_task_weights, colors_hidden_layer, task_layer)

    # Word task process
    comp.add_projection(task_word_weights, task_layer, words_hidden_layer)
    comp.add_projection(word_task_weights, words_hidden_layer, task_layer)

    def trial_dict(red_color, green_color, neutral_color, red_word, green_word, neutral_word, CN, WR):
        trialdict = {
        colors_input_layer: [red_color, green_color, neutral_color],
        words_input_layer: [red_word, green_word, neutral_word],
        task_input_layer: [CN, WR]
        }
        return trialdict

    # Define initialization trials separately
    CN_trial_initialize_input = trial_dict(0, 0, 0, 0, 0, 0, 1, 0)  #red_color, green color, red_word, green word, CN, WR
    CN_incongruent_trial_input = trial_dict(1, 0, 0, 0, 1, 0, 1, 0) #red_color, green color, red_word, green word, CN, WR
    CN_congruent_trial_input = trial_dict(1, 0, 0, 1, 0, 0, 1, 0)   #red_color, green color, red_word, green word, CN, WR
    CN_control_trial_input = trial_dict(1, 0, 0, 0, 0, 1, 1, 0)     #red_color, green color, red_word, green word, CN, WR

    Stimulus = [[CN_trial_initialize_input, CN_congruent_trial_input],
                [CN_trial_initialize_input, CN_incongruent_trial_input],
                [CN_trial_initialize_input, CN_control_trial_input]]

    # should be 500 and 1000
    ntrials0 = 5 * reps
    ntrials = 10 * reps

    def run(bin_execute):
        results = []
        for i, stim in enumerate(Stimulus):
            # RUN the COMPOSITION to initialize --------------------------------
            exec_id = "exec_" + str(i)
            comp.run(inputs=stim[0], num_trials=ntrials0, bin_execute=bin_execute, context=exec_id)
            comp.run(inputs=stim[1], num_trials=ntrials, bin_execute=bin_execute, context=exec_id)

            # Comp results include concatenation of both the above runs
            results.append(comp.results)

        return results

    res = run(mode)
    # the corresponding output port indices in composition results
    # these were 0 and 1 in the prior version of the test
    response_results_index = 3
    response_decision_energy_index = 4
    if reps == 1:
        res2d = [x[response_results_index] for r in res for x in r]
        # NOTE: The formatting below provides visual split between
        #       initialization and runs. Please do not change it.
        assert np.allclose(res2d, [[0.497679878752004, 0.497679878752004],
                                   [0.4954000154631831, 0.4954000154631831],
                                   [0.4931599760310996, 0.4931599760310996],
                                   [0.4909593232354856, 0.4909593232354856],
                                   [0.4887976172454234, 0.4887976172454234],
[0.4866744160981826, 0.4866744160981826],
[0.4845913708928323, 0.4845892761509014],
[0.4825481248817388, 0.4825417411478152],
[0.4805443207349158, 0.4805313535747981],
[0.4785796008355799, 0.4785576550393446],
[0.4766536075535322, 0.4766201866281457],
[0.4747659834976122, 0.4747184892433102],
[0.4729163717484405, 0.4728521039182508],
[0.471104416072617, 0.4710205721142167],
[0.4693297611195044, 0.4692234359984253],
                                   [0.497679878752004, 0.497679878752004],
                                   [0.4954000154631831, 0.4954000154631831],
                                   [0.4931599760310996, 0.4931599760310996],
                                   [0.4909593232354856, 0.4909593232354856],
                                   [0.4887976172454234, 0.4887976172454234],
[0.4866744160981826, 0.4866744160981826],
[0.4845900488307297, 0.484590598212878],
[0.4825440921072959, 0.4825457739209331],
[0.4805361215634117, 0.4805395527400902],
[0.478565712169312, 0.4785715436856292],
[0.4766324385709689, 0.4766413555592825],
[0.4747358754098249, 0.4747485972170732],
[0.4728755976222633, 0.4728928778172651],
[0.4710511807198248, 0.4710738070495766],
[0.4692622010511354, 0.4692909953467661],
                                   [0.497679878752004, 0.497679878752004],
                                   [0.4954000154631831, 0.4954000154631831],
                                   [0.4931599760310996, 0.4931599760310996],
                                   [0.4909593232354856, 0.4909593232354856],
                                   [0.4887976172454234, 0.4887976172454234],
[0.4866744160981826, 0.4866744160981826],
[0.4845900488307297, 0.4845892761509014],
[0.4825440987479238, 0.4825417477884414],
[0.4805361483673633, 0.480531380378737],
[0.4785657797808562, 0.4785577226508331],
[0.4766325749933464, 0.476620323050344],
[0.4747361162403673, 0.4747187300733827],
[0.4728759862850626, 0.4728524925799762],
[0.4710517686957791, 0.4710211600879554],
[0.469263048105203, 0.469224283048266]])

        res1d = [x[response_decision_energy_index] for r in res for x in r]
        assert np.allclose(res1d, [[0.9907410468584376], [0.9816847012836883], [0.9728270478359791], [0.964164228287384], [0.9556924424992138],
[0.9474079491380278], [0.9393111265997224], [0.9313984494721904], [0.9236664515817239], [0.9161117261021626],
[0.9087309255565738], [0.9015207617204032], [0.8944780054345429], [0.8875994863362322], [0.880882092515256],
                                   [0.9907410468584376], [0.9816847012836883], [0.9728270478359791], [0.964164228287384], [0.9556924424992138],
[0.9474079491380278], [0.9393111266035642], [0.9313984495075565], [0.9236664517261579], [0.9161117265115205],
[0.9087309264959722], [0.9015207635977346], [0.8944780088366047], [0.887599492067544], [0.8808821016396065],
                                   [0.9907410468584376], [0.9816847012836883], [0.9728270478359791], [0.964164228287384], [0.9556924424992138],
[0.9474079491380278], [0.939308563971253], [0.9313906911792855], [0.9236507947874026], [0.9160853988421866],
[0.9086910874785843], [0.901464504886388], [0.894402355184426], [0.8875014022102764], [0.8807584692328312]])
    if reps == 10:
        assert np.allclose(res[0][ntrials0 - 1][response_results_index], [0.42481045, 0.42481045])
        assert np.allclose(res[0][-1][response_results_index], [0.43512335, 0.39995991])
        assert np.allclose(res[1][ntrials0 - 1][response_results_index], [0.42481045, 0.42481045])
        assert np.allclose(res[1][-1][response_results_index], [0.41360321, 0.42121262])
        assert np.allclose(res[2][ntrials0 - 1][response_results_index], [0.42481045, 0.42481045])
        assert np.allclose(res[2][-1][response_results_index], [0.41621778, 0.40255998])
        assert np.allclose(res[0][ntrials0 - 1][response_decision_energy_index], [0.72185566])
        assert np.allclose(res[0][-1][response_decision_energy_index], [0.69612758])
        assert np.allclose(res[1][ntrials0 - 1][response_decision_energy_index], [0.72185566])
        assert np.allclose(res[1][-1][response_decision_energy_index], [0.69685957])
        assert np.allclose(res[2][ntrials0 - 1][response_decision_energy_index], [0.72185566])
        assert np.allclose(res[2][-1][response_decision_energy_index], [0.67021047])
    if reps == 100:
        assert np.allclose(res[0][ntrials0 - 1][response_results_index], [0.48590224, 0.48590224])
        assert np.allclose(res[0][-1][response_results_index], [0.95967791, 0.21434208])
        assert np.allclose(res[1][ntrials0 - 1][response_results_index], [0.48590224, 0.48590224])
        assert np.allclose(res[1][-1][response_results_index], [0.55847666, 0.83814112])
        assert np.allclose(res[2][ntrials0 - 1][response_results_index], [0.48590224, 0.48590224])
        assert np.allclose(res[2][-1][response_results_index], [0.89673726, 0.25100269])
        assert np.allclose(res[0][ntrials0 - 1][response_decision_energy_index], [0.94440397])
        assert np.allclose(res[0][-1][response_decision_energy_index], [0.82279743])
        assert np.allclose(res[1][ntrials0 - 1][response_decision_energy_index], [0.94440397])
        assert np.allclose(res[1][-1][response_decision_energy_index], [1.87232903])
        assert np.allclose(res[2][ntrials0 - 1][response_decision_energy_index], [0.94440397])
        assert np.allclose(res[2][-1][response_decision_energy_index], [0.90033387])
    if benchmark.enabled:
        benchmark(run, mode)



from psyneulink import *
import numpy as np

class TestControlMechanism:

    # def test_control_modulation_in_system(self):
    #     Tx = TransferMechanism(name='Tx')
    #     Ty = TransferMechanism(name='Ty')
    #     Tz = TransferMechanism(name='Tz')
    #     C =  ModulatoryMechanism(
    #             # function=Linear,
    #             default_variable=[1],
    #             monitor_for_modulation=Ty,
    #             modulatory_signals=ControlSignal(modulation=OVERRIDE,
    #                                               projections=(SLOPE,Tz)))
    #     P1=Process(pathway=[Tx,Tz])
    #     P2=Process(pathway=[Ty, C])
    #     S=System(processes=[P1, P2])
    #     from pprint import pprint
    #     pprint(S.execution_graph)
    #
    #     assert Tz.parameter_ports[SLOPE].mod_afferents[0].sender.owner == C
    #     result = S.run(inputs={Tx:[1,1], Ty:[4,4]})
    #     assert result == [[[4.], [4.]], [[4.], [4.]]]

    # DEPRECATED FUNCTIONALITY 9/26/19
    # def test_assignment_of_control_and_gating_signals(self):
    #     m = ProcessingMechanism(function=Logistic)
    #     c = ControlMechanism(
    #             control_signals=[
    #                 ControlSignal(name="CS1", modulates=(GAIN, m)),
    #                 GatingSignal(name="GS", modulates=m),
    #                 ControlSignal(name="CS2", modulates=(BIAS, m)),
    #             ]
    #     )
    #     assert  c.output_ports.names == ['CS1', 'GS', 'CS2']
    #     assert m.parameter_ports['gain'].mod_afferents[0].sender.owner == c
    #     assert m.parameter_ports['bias'].mod_afferents[0].sender.owner == c
    #     assert m.input_port.mod_afferents[0].sender.owner == c


    def test_control_modulation_in_composition(self):
        Tx = TransferMechanism(name='Tx')
        Ty = TransferMechanism(name='Ty')
        Tz = TransferMechanism(name='Tz')
        C = ControlMechanism(
                default_variable=[1],
                monitor_for_modulation=Ty,
                control_signals=ControlSignal(modulation=OVERRIDE,
                                                 modulates=(SLOPE, Tz)))

        comp = Composition(enable_controller=True)
        comp.add_linear_processing_pathway(pathway=[Tx,Tz])
        comp.add_node(Ty, required_roles=NodeRole.TERMINAL)
        comp.add_controller(C)

        assert Tz.parameter_ports[SLOPE].mod_afferents[0].sender.owner == C
        assert np.allclose(comp.results,[[[1.], [4.]], [[4.], [4.]]])



from psyneulink import *
from gym_forager.envs.forager_env import ForagerEnv
from gym_forager.envs.forager.randomexp import RandomAgent


num_trials = 4
env = ForagerEnv()
reward = 0
done = False


# Function used by PsyNeuLink Mechanism
r = RandomAgent(env.action_space)
def random_action(variable):
    return r.act(variable, None, None)

# Create PsyNeuLink Composition as agent
agent_mech = ProcessingMechanism(function=random_action)
agent_comp = Composition()

agent_comp.add_node(agent_mech)

def main():
    for _ in range(num_trials):
        observation = env.reset()
        while True:
            run_results = agent_comp.run(inputs={agent_mech:observation})
            action=run_results[0]
            observation, reward, done, _ = env.step(action)
            if done:
                break

if __name__ == "__main__":
    main()



from . import controlprojection
from . import gatingprojection
from . import learningprojection
from . import modulatoryprojection

from .controlprojection import *
from .gatingprojection import *
from .learningprojection import *
from .modulatoryprojection import *

__all__ = list(controlprojection.__all__)
__all__.extend(gatingprojection.__all__)
__all__.extend(learningprojection.__all__)
__all__.extend(modulatoryprojection.__all__)



from psyneulink.core import llvm as pnlvm
from psyneulink.library.compositions.pytorchllvmhelper import *

__all__ = ['MSELoss']


class Loss():

    def __init__(self):
        self._DELTA_W_NUM = 0


    def _gen_llvm_function(self, *, ctx:pnlvm.LLVMBuilderContext, tags:frozenset):
        return self._gen_loss_function(ctx)


    def gen_inject_lossfunc_call(self, ctx, builder, bin_func, value, target):
        return builder.call(bin_func, [builder.gep(value, [ctx.int32_ty(0), ctx.int32_ty(0)]),
                                       ctx.int32_ty(len(value.type.pointee)),
                                       builder.gep(target, [ctx.int32_ty(0), ctx.int32_ty(0)])])

class MSELoss(Loss):
    """Implements compiled MSE Loss"""
    def __init__(self, reduction='sum'):
        if reduction not in ['sum']:
            raise Exception("Unsupported compiled reduction type " + reduction)

        super().__init__()
        self.reduction = reduction

    def _gen_loss_function(self, ctx):
        name = "LEARNING_MSE_CALL"

        # args:
        # 1) pointer to network output
        # 2) pointer to target
        # 3) dimensionality
        args = [ctx.float_ty.as_pointer(), ctx.int32_ty, ctx.float_ty.as_pointer()]
        builder = ctx.create_llvm_function(args, self, name, return_type=ctx.float_ty)
        value, dim, target = builder.function.args

        sum = builder.alloca(ctx.float_ty)
        builder.store(ctx.float_ty(-0.0), sum)

        with pnlvm.helpers.for_loop_zero_inc(builder, dim, "mse_sum_loop") as (b1, index):
            value_ptr = b1.gep(value,[index])
            target_ptr = b1.gep(target,[index])
            diff = b1.fsub(b1.load(value_ptr),b1.load(target_ptr))
            diff = b1.fmul(diff,diff)
            b1.store(b1.fadd(b1.load(sum),diff),sum)

        # Average the values in sum by dimensionality
        builder.store(builder.fdiv(builder.load(sum),builder.uitofp(dim, ctx.float_ty)), sum)

        builder.ret(builder.load(sum))

        return builder.function

    # inserts the computation for dC/da
    def _gen_inject_loss_differential(self, ctx, builder, value, target, output=None, sum_loss=False):
        dim = len(value.type.pointee)
        assert len(target.type.pointee) == dim
        if output is None:
            output = builder.alloca(pnlvm.ir.types.ArrayType(ctx.float_ty, dim))
            # zero output vector
            builder.store(output.type.pointee(None), output)
        assert len(output.type.pointee) == dim

        if sum_loss is False:
            # we take mean
            gen_inject_vec_sub(ctx, builder, value, target, output)
            # multiply each element i by 2/n to get dC/da_i
            scalar_mult = builder.fdiv(ctx.float_ty(2), ctx.float_ty(dim))
            with pnlvm.helpers.for_loop_zero_inc(builder, ctx.int32_ty(dim), "mse_mean_mult_loop") as (b1, index):
                element_ptr = b1.gep(output, [ctx.int32_ty(0), index])
                b1.store(b1.fmul(b1.load(element_ptr),scalar_mult),element_ptr)
        else:
            # in this case, we add the loss
            tmp = gen_inject_vec_sub(ctx, builder, value, target)
            gen_inject_vec_add(ctx, builder, output, tmp, output)
        return output



import numpy as np
import psyneulink as pnl
import pytest

import psyneulink.core.components.functions.transferfunctions


@pytest.mark.parametrize(
    'output_type, variable, expected_output',
    [
        (pnl.FunctionOutputType.RAW_NUMBER, 1, 1.),
        (pnl.FunctionOutputType.RAW_NUMBER, [1], 1.),
        (pnl.FunctionOutputType.RAW_NUMBER, [[1]], 1.),
        (pnl.FunctionOutputType.RAW_NUMBER, [[[1]]], 1.),
        (pnl.FunctionOutputType.NP_1D_ARRAY, 1, np.array([1.])),
        (pnl.FunctionOutputType.NP_1D_ARRAY, [1], np.array([1.])),
        (pnl.FunctionOutputType.NP_1D_ARRAY, [[1]], np.array([1.])),
        (pnl.FunctionOutputType.NP_1D_ARRAY, [[[1]]], np.array([1.])),
        (pnl.FunctionOutputType.NP_2D_ARRAY, 1, np.array([[1.]])),
        (pnl.FunctionOutputType.NP_2D_ARRAY, [1], np.array([[1.]])),
        (pnl.FunctionOutputType.NP_2D_ARRAY, [[1]], np.array([[1.]])),
        (pnl.FunctionOutputType.NP_2D_ARRAY, [[[1]]], np.array([[1.]])),
    ]
)
def test_output_type_conversion(output_type, variable, expected_output):
    f = psyneulink.core.components.functions.transferfunctions.Linear()
    f.output_type = output_type
    f.enable_output_type_conversion = True

    assert f.execute(variable) == expected_output


@pytest.mark.parametrize(
    'output_type, variable',
    [
        (pnl.FunctionOutputType.RAW_NUMBER, [1, 1]),
        (pnl.FunctionOutputType.RAW_NUMBER, [[1, 1]]),
        (pnl.FunctionOutputType.RAW_NUMBER, [[[1], [1, 1]]]),
        (pnl.FunctionOutputType.NP_1D_ARRAY, [[1, 1], [1, 1]]),
    ]
)
def test_output_type_conversion_failure(output_type, variable):
    with pytest.raises(pnl.FunctionError):
        f = psyneulink.core.components.functions.transferfunctions.Linear()
        f.convert_output_type(variable, output_type=output_type)



import ctypes
import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm

ITERATIONS=100
DIM_X=1000
DIM_Y=2000

matrix = np.random.rand(DIM_X, DIM_Y)
vector = np.random.rand(DIM_X)
llvm_res = np.random.rand(DIM_Y)

ct_vec = vector.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
ct_mat = matrix.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
x, y = matrix.shape

@pytest.mark.llvm
def test_recompile():
    # The original builtin mxv function
    binf = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_vxm')
    orig_res = np.empty_like(llvm_res)
    ct_res = orig_res.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    binf.c_func(ct_vec, ct_mat, x, y, ct_res)

    # Rebuild and try again
    # This is not a public API
    pnlvm._llvm_build()

    rebuild_res = np.empty_like(llvm_res)
    ct_res = rebuild_res.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    binf.c_func(ct_vec, ct_mat, x, y, ct_res)
    assert np.array_equal(orig_res, rebuild_res)

    # Get a new pointer
    binf2 = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_vxm')
    new_res = np.empty_like(llvm_res)
    ct_res = new_res.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    binf.c_func(ct_vec, ct_mat, x, y, ct_res)
    assert np.array_equal(rebuild_res, new_res)

    callable_res = np.empty_like(llvm_res)
    ct_res = callable_res.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    binf(ct_vec, ct_mat, x, y, ct_res)
    assert np.array_equal(new_res, callable_res)
